"""
===============================
05. Simulate beta modulated ERP
===============================

This example demonstrates how event related potentials (ERP) are modulated
by prestimulus beta events. Specifically, this example reproduces Figure 5
from Law et al. 2021 [1]_. To be consistent with the publication, the default
network connectivity is altered. These modifications demonstrate a potential
mechanism by which transient beta activity in the neocortex can suppress
the perceptibility of sensory input. This suppression depends on the timing
of the beta event, and the incoming sensory information.
"""

# Authors: Nick Tolley <nicholas_tolley@brown.edu>

from hnn_core import simulate_dipole, law_2021_model, jones_2009_model
from hnn_core.viz import plot_dipole

###############################################################################
# We begin by instantiating the network model from Law et al. 2021 [1]_.
net = law_2021_model()

###############################################################################
# The Law 2021 model is based on the network model described in
# Jones et al. 2009 [2]_ with several important modifications. One of the most
# significant changes is substantially increasing the rise and fall time
# constants of GABAb-conductances on L2 and L5 pyramidal. Another important
# change is the removal of calcium channels from basal dendrites and soma of
# L5 pyramidal cells specifically.
# We can inspect these properties with the ``net.cell_types`` attribute which
# contains information on the biophysics and geometry of each cell.
net_jones = jones_2009_model()

jones_rise = net_jones.cell_types['L5_pyramidal']['cell_object'].synapses['gabab']['tau1']
law_rise = net.cell_types['L5_pyramidal']['cell_object'].synapses['gabab']['tau1']
print(f'GABAb Rise (ms): {jones_rise} -> {law_rise}')

jones_fall = net_jones.cell_types['L5_pyramidal']['cell_object'].synapses['gabab']['tau2']
law_fall = net.cell_types['L5_pyramidal']['cell_object'].synapses['gabab']['tau2']
print(f'GABAb Fall (ms): {jones_fall} -> {law_fall}\n')

print('Apical Dendrite Channels:')
print(net.cell_types['L5_pyramidal']['cell_object'].sections['apical_1'].mechs.keys())
print("\nBasal Dendrite Channels ('ca' missing):")
print(net.cell_types['L5_pyramidal']['cell_object'].sections['basal_1'].mechs.keys())

###############################################################################
# A major change to the Jones 2009 model is the addition of a
# Martinotti-like recurrent tuft connection [3]_. This new connection
# originates from L5 basket cells, and provides GABAa inhibition on
# the distal dendrites of L5 pyramidal cells.
print('Recurrent Tuft Connection')
print(net.connectivity[16])

###############################################################################
# The remaining changes to the connectivity was the removal of an
# L2_basket -> L5_pyramidal GABAa connection, and replacing it with GABAb.
print('New GABAb connection')
print(net.connectivity[15])

print('\nConnection Removed from Law Model')
print(net_jones.connectivity[10])


###############################################################################
# To demonstrate sensory depression, we will add the drives necessary to
# generate and ERP similar to
# :ref:`evoked example <sphx_glr_auto_examples_plot_simulate_evoked.py>`,
# but modified to reflect the parameters used in Law et al. 2021.
# Specifically, we are considering the case where a tactile stimulus is
# delivered at 150 ms. 25 ms later, the first input to sensory cortex arrives
# as a proximal drive to the cortical column. Proximal drive corresponds to
# projections from the direct thalamic nuclei. This is followed by one
# :term:`distal` drive representing projections from indirect thalamic nuclei,
# and a final late proximal drive. It is important to note that the parameter
# values for each are different from previous examples of the evoked response.
# This reflects the altered network dynamics due to the changes described
# above.
def add_erp_drives(net, stimulus_start):
    # Distal evoked drive
    weights_ampa_d1 = {'L2_basket': 0.0005, 'L2_pyramidal': 0.004,
                       'L5_pyramidal': 0.0005}
    weights_nmda_d1 = {'L2_basket': 0.0005, 'L2_pyramidal': 0.004,
                       'L5_pyramidal': 0.0005}
    syn_delays_d1 = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                     'L5_pyramidal': 0.1}
    net.add_evoked_drive(
        'evdist1', mu=70.0 + stimulus_start, sigma=0.0, numspikes=1,
        weights_ampa=weights_ampa_d1, weights_nmda=weights_nmda_d1,
        location='distal', synaptic_delays=syn_delays_d1, event_seed=274)

    # Two proximal drives
    weights_ampa_p1 = {'L2_basket': 0.002, 'L2_pyramidal': 0.0011,
                       'L5_basket': 0.001, 'L5_pyramidal': 0.001}
    syn_delays_prox = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                       'L5_basket': 1., 'L5_pyramidal': 1.}

    # all NMDA weights are zero; pass None explicitly
    net.add_evoked_drive(
        'evprox1', mu=25.0 + stimulus_start, sigma=0.0, numspikes=1,
        weights_ampa=weights_ampa_p1, weights_nmda=None,
        location='proximal', synaptic_delays=syn_delays_prox, event_seed=544)

    # Second proximal evoked drive. NB: only AMPA weights differ from first
    weights_ampa_p2 = {'L2_basket': 0.005, 'L2_pyramidal': 0.005,
                       'L5_basket': 0.01, 'L5_pyramidal': 0.01}
    # all NMDA weights are zero; omit weights_nmda (defaults to None)
    net.add_evoked_drive(
        'evprox2', mu=135.0 + stimulus_start, sigma=0.0, numspikes=1,
        weights_ampa=weights_ampa_p2, location='proximal',
        synaptic_delays=syn_delays_prox, event_seed=814)

    return net


###############################################################################
# A beta event is created by inducing simultaneous proximal and distal
# drives. The input is just strong enough to evoke spiking in the
# L2 basket cells. This spiking causes GABAb mediated inhibition
# of the network, and ultimately suppressed sensory detection.
def add_beta_drives(net, beta_start):
    # Distal Drive
    weights_ampa_d1 = {'L2_basket': 0.00032, 'L2_pyramidal': 0.00008,
                       'L5_pyramidal': 0.00004}
    syn_delays_d1 = {'L2_basket': 0.5, 'L2_pyramidal': 0.5,
                     'L5_pyramidal': 0.5}
    net.add_bursty_drive(
        'beta_dist', tstart=beta_start, tstart_std=0., tstop=beta_start + 50.,
        burst_rate=1., burst_std=10., numspikes=2, spike_isi=10,
        n_drive_cells=10, location='distal', weights_ampa=weights_ampa_d1,
        synaptic_delays=syn_delays_d1, event_seed=290)

    # Proximal Drive
    weights_ampa_p1 = {'L2_basket': 0.00004, 'L2_pyramidal': 0.00002,
                       'L5_basket': 0.00002, 'L5_pyramidal': 0.00002}
    syn_delays_p1 = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                     'L5_basket': 1.0, 'L5_pyramidal': 1.0}

    net.add_bursty_drive(
        'beta_prox', tstart=beta_start, tstart_std=0., tstop=beta_start + 50.,
        burst_rate=1., burst_std=20., numspikes=2, spike_isi=10,
        n_drive_cells=10, location='proximal', weights_ampa=weights_ampa_p1,
        synaptic_delays=syn_delays_p1, event_seed=300)
    return net


###############################################################################
# We can now use our functions to create three distinct simulations:
# 1) beta event only, 2) ERP only, and 3) beta event + ERP.
beta_start, stimulus_start = 50.0, 125.0
net_beta = net.copy()
net_beta = add_beta_drives(net_beta, beta_start)

net_erp = net.copy()
net_erp = add_erp_drives(net_erp, stimulus_start)

net_beta_erp = net_beta.copy()
net_beta_erp = add_erp_drives(net_beta_erp, stimulus_start)

###############################################################################
# And finally we simulate. Note that the default simulation time has been
# increased to 400 ms to observe the long time course over which beta events
# can influence sensory input to the cortical column.
dpls_beta = simulate_dipole(net_beta, tstop=400)
dpls_erp = simulate_dipole(net_erp, tstop=400)
dpls_beta_erp = simulate_dipole(net_beta_erp, tstop=400)

###############################################################################
# By inspecting the activity during the beta event, we can see that spiking
# occurs exclusively at 50 ms, the peak of the gaussian distributed proximal
# and distal inputs. This spiking activity leads to sustained GABAb mediated
# inhibition of the L2 and L5 pyrmaidal cells. One effect of this inhibition
# is an asymmetric beta event with a long positive tail.
import matplotlib.pyplot as plt
import numpy as np
fig, axes = plt.subplots(4, 1, sharex=True, figsize=(7, 7),
                         constrained_layout=True)
net_beta.cell_response.plot_spikes_hist(ax=axes[0], show=False)
axes[0].set_title('Beta Event Generation')
plot_dipole(dpls_beta, ax=axes[1], layer='agg', tmin=1.0, color='b', show=False)
net_beta.cell_response.plot_spikes_raster(ax=axes[2], show=False)
axes[2].set_title('Spike Raster')

# Create a fixed-step tiling of frequencies from 1 to 40 Hz in steps of 1 Hz
freqs = np.arange(10., 60., 1.)
dpls_beta[0].plot_tfr_morlet(freqs, n_cycles=7, ax=axes[3])

###############################################################################
# Next we will inspect what happens when a sensory stimulus is delivered 75 ms
# after a beta event. Note that the delay time for a tactile stimulus at the
# hand to arrive at the cortex is roughly 25 ms, which means the first proximal
# input to the cortical column occurs ~100 ms after the beta event.
dpls_beta_erp[0].smooth(45)
fig, axes = plt.subplots(3, 1, sharex=True, figsize=(7, 7),
                         constrained_layout=True)
plot_dipole(dpls_beta_erp, ax=axes[0], layer='agg', tmin=1.0, color='r',
            show=False)
axes[0].set_title('Beta Event + ERP')
net_beta_erp.cell_response.plot_spikes_hist(ax=axes[1], show=False)
axes[1].set_title('Input Drives Histogram')
net_beta_erp.cell_response.plot_spikes_raster(ax=axes[2], show=False)
axes[2].set_title('Spike Raster')

###############################################################################
# To help understand the effect of beta mediated inhibition on the response to
# incoming sensory stimuli, we can compare the ERP and spiking activity due to
# sensory input with and without a beta event.
# The sustained inhibition of the network ultimately depresses
# the sensory response which is associated with a reduced ERP amplitude
dpls_erp[0].smooth(45)
fig, axes = plt.subplots(3, 1, sharex=True, figsize=(7, 7),
                         constrained_layout=True)
plot_dipole(dpls_beta_erp, ax=axes[0], layer='agg', tmin=1.0, color='r',
            show=False)
plot_dipole(dpls_erp, ax=axes[0], layer='agg', tmin=1.0, color='b', show=False)
axes[0].set_title('Beta ERP Comparison')
axes[0].legend(['ERP + Beta', 'ERP'])
net_beta_erp.cell_response.plot_spikes_raster(ax=axes[1], show=False)
axes[1].set_title('Beta + ERP Spike Raster')
net_erp.cell_response.plot_spikes_raster(ax=axes[2], show=False)
axes[2].set_title('ERP Spike Raster')
plt.show()

###############################################################################
# References
# ----------
# .. [1] Law, R. G., Pugliese, S., Shin, H., Sliva, D. D., Lee, S.,
#        Neymotin, S., Moore, C., & Jones, S. R. (2021). Thalamocortical
#        mechanisms regulating the relationship between transient beta events
#        and human tactile perception. BioRxiv, 2021.04.16.440210.
#        https://doi.org/10.1101/2021.04.16.440210
# .. [2] Jones, S. R., Pritchett, D. L., Sikora, M. A., Stufflebeam, S. M.,
#        Hämäläinen, M., & Moore, C. I. (2009). Quantitative Analysis and
#        Biophysically Realistic Neural Modeling of the MEG Mu Rhythm:
#        Rhythmogenesis and Modulation of Sensory-Evoked Responses. Journal of
#        Neurophysiology, 102(6), 3554–3572.
#        https://doi.org/10.1152/jn.00535.2009
# .. [3] Silberberg, G., & Markram, H. (2007). Disynaptic Inhibition between
#        Neocortical Pyramidal Cells Mediated by Martinotti Cells. Neuron,
#        53(5), 735–746. https://doi.org/10.1016/j.neuron.2007.02.012



import os.path as op

import matplotlib
from matplotlib import backend_bases
import matplotlib.pyplot as plt
from matplotlib.colorbar import Colorbar

import numpy as np
from numpy.testing import assert_allclose
import pytest

import hnn_core
from hnn_core import read_params, jones_2009_model
from hnn_core.viz import (
    plot_cells,
    plot_dipole,
    plot_psd,
    plot_tfr_morlet,
    plot_connectivity_matrix,
    plot_cell_connectivity,
    plot_drive_strength,
    NetworkPlotter,
)
from hnn_core.dipole import simulate_dipole

matplotlib.use("agg")


@pytest.fixture(autouse=True)
def cleanup_matplotlib():
    # Code runs after the test finishes
    yield
    plt.close("all")


@pytest.fixture
def setup_net():
    hnn_core_root = op.dirname(hnn_core.__file__)
    params_fname = op.join(hnn_core_root, "param", "default.json")
    params = read_params(params_fname)
    net = jones_2009_model(params, mesh_shape=(3, 3))

    return net


def _fake_click(fig, ax, point, button=1):
    """Fake a click at a point within axes."""
    x, y = ax.transData.transform_point(point)
    button_press_event = backend_bases.MouseEvent(
        name="button_press_event", canvas=fig.canvas, x=x, y=y, button=button
    )
    fig.canvas.callbacks.process("button_press_event", button_press_event)


def test_network_visualization(setup_net):
    """Test network visualisations."""
    net = setup_net
    plot_cells(net)
    ax = net.cell_types["L2_pyramidal"]["cell_object"].plot_morphology()
    assert len(ax.lines) == 8

    conn_idx = 0
    plot_connectivity_matrix(net, conn_idx, show=False)
    with pytest.raises(TypeError, match="net must be an instance of"):
        plot_connectivity_matrix("blah", conn_idx, show_weight=False)

    with pytest.raises(TypeError, match="conn_idx must be an instance of"):
        plot_connectivity_matrix(net, "blah", show_weight=False)

    with pytest.raises(TypeError, match="show_weight must be an instance of"):
        plot_connectivity_matrix(net, conn_idx, show_weight="blah")

    src_gid = 5
    plot_cell_connectivity(net, conn_idx, src_gid, show=False)
    with pytest.raises(TypeError, match="net must be an instance of"):
        plot_cell_connectivity("blah", conn_idx, src_gid=src_gid)

    with pytest.raises(TypeError, match="conn_idx must be an instance of"):
        plot_cell_connectivity(net, "blah", src_gid)

    with pytest.raises(TypeError, match="src_gid must be an instance of"):
        plot_cell_connectivity(net, conn_idx, src_gid="blah")

    with pytest.raises(ValueError, match="src_gid -1 not a valid cell ID"):
        plot_cell_connectivity(net, conn_idx, src_gid=-1)

    # Test morphology plotting
    for cell_type in net.cell_types.values():
        cell_type["cell_object"].plot_morphology()
        cell_type["cell_object"].plot_morphology(color="r")

        # FIX: Access sections through the cell object
        sections = list(cell_type["cell_object"].sections.keys())
        section_color = {sect_name: f"C{idx}" for idx, sect_name in enumerate(sections)}
        cell_type["cell_object"].plot_morphology(color=section_color)

    cell_type = net.cell_types["L2_basket"]
    with pytest.raises(ValueError):
        cell_type["cell_object"].plot_morphology(color="z")
    with pytest.raises(ValueError):
        cell_type["cell_object"].plot_morphology(color={"soma": "z"})
    with pytest.raises(TypeError, match="color must be"):
        cell_type["cell_object"].plot_morphology(color=123)

    # test for invalid Axes object to plot_cells
    fig, axes = plt.subplots(1, 1)
    with pytest.raises(
        TypeError, match="'ax' to be an instance of Axes3D, but got Axes"
    ):
        plot_cells(net, ax=axes, show=False)
    cell_type["cell_object"].plot_morphology(pos=(1.0, 2.0, 3.0))
    with pytest.raises(TypeError, match="pos must be"):
        cell_type["cell_object"].plot_morphology(pos=123)
    with pytest.raises(ValueError, match="pos must be a tuple of 3 elements"):
        cell_type["cell_object"].plot_morphology(pos=(1, 2, 3, 4))
    with pytest.raises(TypeError, match="pos\\[idx\\] must be"):
        cell_type["cell_object"].plot_morphology(pos=(1, "2", 3))

    plt.close("all")

    # test interactive clicking updates the position of src_cell in plot
    del net.connectivity[-1]
    conn_idx = 15
    net.add_connection(
        net.gid_ranges["L2_pyramidal"][::2],
        "L5_basket",
        "soma",
        "ampa",
        0.00025,
        1.0,
        lamtha=3.0,
        probability=0.8,
    )
    fig = plot_cell_connectivity(net, conn_idx, show=False)
    ax_src, ax_target, _ = fig.axes

    pos = net.pos_dict["L2_pyramidal"][2]
    _fake_click(fig, ax_src, [pos[0], pos[1]])
    pos_in_plot = ax_target.collections[2].get_offsets().data[0]
    assert_allclose(pos[:2], pos_in_plot)
    plt.close("all")


def test_dipole_visualization(setup_net):
    """Test dipole visualisations."""
    net = setup_net

    # Test plotting of simulations with no spiking
    dpls = simulate_dipole(net, tstop=100.0, n_trials=1)
    net.cell_response.plot_spikes_raster()
    net.cell_response.plot_spikes_hist()

    weights_ampa = {"L2_pyramidal": 5.4e-5, "L5_pyramidal": 5.4e-5}
    syn_delays = {"L2_pyramidal": 0.1, "L5_pyramidal": 1.0}

    net.add_bursty_drive(
        "beta_prox",
        tstart=0.0,
        burst_rate=25,
        burst_std=5,
        numspikes=1,
        spike_isi=0,
        n_drive_cells=11,
        location="proximal",
        weights_ampa=weights_ampa,
        synaptic_delays=syn_delays,
        event_seed=14,
    )

    net.add_bursty_drive(
        "beta_dist",
        tstart=0.0,
        burst_rate=25,
        burst_std=5,
        numspikes=1,
        spike_isi=0,
        n_drive_cells=11,
        location="distal",
        weights_ampa=weights_ampa,
        synaptic_delays=syn_delays,
        event_seed=14,
    )

    dpls = simulate_dipole(net, tstop=100.0, n_trials=2, record_vsec="all")
    fig = dpls[0].plot()  # plot the first dipole alone
    axes = fig.get_axes()[0]
    dpls[0].copy().smooth(window_len=10).plot(ax=axes)  # add smoothed versions
    dpls[0].copy().savgol_filter(h_freq=30).plot(ax=axes)  # on top

    # test decimation options
    plot_dipole(dpls[0], decim=2, show=False)
    for dec in [-1, [2, 2.0]]:
        with pytest.raises(
            ValueError, match="each decimation factor must be a positive"
        ):
            plot_dipole(dpls[0], decim=dec, show=False)

    # test plotting multiple dipoles as overlay
    fig = plot_dipole(dpls, show=False)

    # test plotting multiple dipoles with average
    fig = plot_dipole(dpls, average=True, show=False)
    plt.close("all")

    # test plotting dipoles with multiple layers
    fig, ax = plt.subplots()
    fig = plot_dipole(dpls, show=False, ax=[ax], layer=["L2"])
    fig = plot_dipole(dpls, show=False, layer=["L2", "L5", "agg"])
    fig, axes = plt.subplots(nrows=3, ncols=1)
    fig = plot_dipole(dpls, show=False, ax=axes, layer=["L2", "L5", "agg"])
    fig, axes = plt.subplots(nrows=3, ncols=1)
    fig = plot_dipole(
        dpls, show=False, ax=[axes[0], axes[1], axes[2]], layer=["L2", "L5", "agg"]
    )

    plt.close("all")

    with pytest.raises(AssertionError, match="ax and layer should have the same size"):
        fig, axes = plt.subplots(nrows=3, ncols=1)
        fig = plot_dipole(dpls, show=False, ax=axes, layer=["L2", "L5"])

    # multiple TFRs get averaged
    fig = plot_tfr_morlet(dpls, freqs=np.arange(23, 26, 1.0), n_cycles=3, show=False)
    # when min_freq > max_freq (y-axis inversion)
    fig = plot_tfr_morlet(dpls, freqs=np.array([30, 20, 10]), n_cycles=3, show=False)
    ax = fig.get_axes()[0]
    y_limits = ax.get_ylim()
    assert y_limits[0] > y_limits[1], (
        "Y-axis should be inverted when min_freq > max_freq"
    )

    with pytest.raises(RuntimeError, match="All dipoles must be scaled equally!"):
        plot_dipole([dpls[0].copy().scale(10), dpls[1].copy().scale(20)])
    with pytest.raises(RuntimeError, match="All dipoles must be scaled equally!"):
        plot_psd([dpls[0].copy().scale(10), dpls[1].copy().scale(20)])
    with pytest.raises(RuntimeError, match="All dipoles must be sampled equally!"):
        dpl_sfreq = dpls[0].copy()
        dpl_sfreq.sfreq /= 10
        plot_psd([dpls[0], dpl_sfreq])

    # pytest deprecation warning for tmin and tmax
    with pytest.deprecated_call():
        plot_dipole(dpls[0], show=False, tmin=10, tmax=100)

    # test cell response plotting
    with pytest.raises(TypeError, match="trial_idx must be an instance of"):
        net.cell_response.plot_spikes_raster(trial_idx="blah", show=False)
    net.cell_response.plot_spikes_raster(trial_idx=0, show=False)
    fig = net.cell_response.plot_spikes_raster(trial_idx=[0, 1], show=False)
    assert len(fig.axes[0].collections) > 0, "No data plotted in raster plot"

    with pytest.raises(TypeError, match="trial_idx must be an instance of"):
        net.cell_response.plot_spikes_hist(trial_idx="blah")
    net.cell_response.plot_spikes_hist(trial_idx=0, show=False)
    net.cell_response.plot_spikes_hist(trial_idx=[0, 1], show=False)
    net.cell_response.plot_spikes_hist(color="r")
    net.cell_response.plot_spikes_hist(color=["C0", "C1"])
    net.cell_response.plot_spikes_hist(color={"beta_prox": "r", "beta_dist": "g"})
    net.cell_response.plot_spikes_hist(
        spike_types={"group1": ["beta_prox", "beta_dist"]}, color={"group1": "r"}
    )
    net.cell_response.plot_spikes_hist(
        spike_types={"group1": ["beta"]}, color={"group1": "r"}
    )

    with pytest.raises(TypeError, match="color must be an instance of"):
        net.cell_response.plot_spikes_hist(color=123)
    with pytest.raises(ValueError):
        net.cell_response.plot_spikes_hist(color="z")
    with pytest.raises(ValueError):
        net.cell_response.plot_spikes_hist(color={"beta_prox": "z", "beta_dist": "g"})
    with pytest.raises(TypeError, match="Dictionary values of color must"):
        net.cell_response.plot_spikes_hist(color={"beta_prox": 123, "beta_dist": "g"})
    with pytest.raises(ValueError, match="'beta_dist' must be"):
        net.cell_response.plot_spikes_hist(color={"beta_prox": "r"})
    plt.close("all")


def test_drive_strength(setup_net):
    """Adds empty external drives to check there strength across each cell types"""
    net = setup_net

    weights_ampa = {"L2_pyramidal": 0.0, "L5_pyramidal": 0.0, "L2_basket": 0.0}
    synaptic_delays = {"L2_pyramidal": 0.0, "L5_pyramidal": 0.0, "L2_basket": 0.0}
    rate_constant = {"L2_pyramidal": 140.0, "L5_pyramidal": 40.0, "L2_basket": 100.0}

    net.add_poisson_drive(
        "poisson",
        rate_constant=rate_constant,
        weights_ampa=weights_ampa,
        location="proximal",
        synaptic_delays=synaptic_delays,
        event_seed=1349,
    )

    net.add_bursty_drive(
        "beta_dist",
        tstart=0.0,
        burst_rate=25,
        burst_std=5,
        numspikes=1,
        spike_isi=0,
        n_drive_cells=11,
        location="distal",
        weights_ampa=weights_ampa,
        synaptic_delays=synaptic_delays,
        event_seed=14,
    )

    figure = plot_drive_strength(net)

    assert isinstance(figure, plt.Figure)
    assert len(figure.axes) > 0  # if there are any axes in the figure

    any_plot = any(ax.lines or ax.patches or ax.images for ax in figure.axes)
    assert any_plot  # At least one axis contains graphical elements

    plt.close("all")


class TestCellResponsePlotters:
    """Tests plotting methods of the CellResponse class"""

    @pytest.fixture(scope="class")
    def class_setup_net(self):
        """Creates a base network for tests within this class"""
        hnn_core_root = op.dirname(hnn_core.__file__)
        params_fname = op.join(hnn_core_root, "param", "default.json")
        params = read_params(params_fname)
        net = jones_2009_model(params, mesh_shape=(3, 3))

        return net

    @pytest.fixture(scope="class")
    def base_simulation_spikes(self, class_setup_net):
        """Adds drives with spikes for testing of spike visualizations"""
        net = class_setup_net
        weights_ampa = {"L2_pyramidal": 0.1, "L5_pyramidal": 1.0}
        syn_delays = {"L2_pyramidal": 0.1, "L5_pyramidal": 1.0}
        net.add_bursty_drive(
            "beta_prox",
            tstart=0.0,
            burst_rate=25,
            burst_std=5,
            numspikes=1,
            spike_isi=0,
            n_drive_cells=11,
            location="proximal",
            weights_ampa=weights_ampa,
            synaptic_delays=syn_delays,
            event_seed=14,
        )

        net.add_bursty_drive(
            "beta_dist",
            tstart=0.0,
            burst_rate=25,
            burst_std=5,
            numspikes=1,
            spike_isi=0,
            n_drive_cells=11,
            location="distal",
            weights_ampa=weights_ampa,
            synaptic_delays=syn_delays,
            event_seed=14,
        )
        dpls = simulate_dipole(net, tstop=100.0, n_trials=2, record_vsec="all")

        return net, dpls

    def test_spikes_raster_trial_idx(self, base_simulation_spikes):
        """Plotting with different index arguments"""
        net, _ = base_simulation_spikes

        # Bad index argument raises error
        with pytest.raises(TypeError, match="trial_idx must be an instance of"):
            net.cell_response.plot_spikes_raster(trial_idx="blah", show=False)

        # Test valid index arguments
        for index_arg in (0, [0, 1]):
            fig = net.cell_response.plot_spikes_raster(trial_idx=index_arg, show=False)
            # Check that collections contain data
            assert all(
                [
                    collection.get_positions() != [-1]
                    for collection in fig.axes[0].collections
                ]
            ), "No data plotted in raster plot"

    def test_spikes_raster_colors(self, base_simulation_spikes):
        """Plotting with different color arguments"""
        net, _ = base_simulation_spikes

        def _get_line_hex_colors(fig):
            colors = [
                matplotlib.colors.to_hex(line.get_color())
                for line in fig.axes[0].legend_.get_lines()
            ]
            labels = [text.get_text() for text in fig.axes[0].legend_.get_texts()]
            return colors, labels

        # Default colors should be the default color cycle
        fig = net.cell_response.plot_spikes_raster(trial_idx=0, show=False)
        colors, _ = _get_line_hex_colors(fig)
        default_colors = plt.rcParams["axes.prop_cycle"].by_key()["color"][
            0 : len(colors)
        ]
        assert colors == default_colors

        # Custom hex colors as list
        custom_colors = ["#daf7a6", "#ffc300", "#ff5733", "#c70039"]
        fig = net.cell_response.plot_spikes_raster(
            trial_idx=0, show=False, colors=custom_colors
        )
        colors, _ = _get_line_hex_colors(fig)
        assert colors == custom_colors

        # Custom named colors as list
        custom_colors = ["skyblue", "maroon", "gold", "hotpink"]
        color_map = matplotlib.colors.get_named_colors_mapping()
        fig = net.cell_response.plot_spikes_raster(
            trial_idx=0, show=False, colors=custom_colors
        )
        colors, _ = _get_line_hex_colors(fig)
        assert colors == [color_map[color].lower() for color in custom_colors]

        # Incorrect number of colors as list
        too_few = ["r", "g", "b"]
        too_many = ["r", "g", "b", "y", "k"]
        for colors in [too_few, too_many]:
            with pytest.raises(ValueError, match="Number of colors must be equal to"):
                net.cell_response.plot_spikes_raster(
                    trial_idx=0, show=False, colors=colors
                )

        # Colors as dict mapping
        dict_mapping = {
            "L2_basket": "#daf7a6",
            "L2_pyramidal": "#ffc300",
            "L5_basket": "#ff5733",
            "L5_pyramidal": "#c70039",
        }
        fig = net.cell_response.plot_spikes_raster(
            trial_idx=0, show=False, colors=dict_mapping
        )
        colors, _ = _get_line_hex_colors(fig)
        assert colors == list(dict_mapping.values())

        # Change color of only one cell type
        dict_mapping = {"L2_pyramidal": "#daf7a6"}
        fig = net.cell_response.plot_spikes_raster(
            trial_idx=0, show=False, colors=dict_mapping
        )
        colors, cell_types = _get_line_hex_colors(fig)
        assert colors[cell_types.index("L2_pyramidal Spikes")] == "#daf7a6"

        # Invalid key in dict mapping
        dict_mapping = {"bad_cell_type": "#daf7a6"}
        with pytest.raises(ValueError, match="Invalid cell types provided."):
            net.cell_response.plot_spikes_raster(
                trial_idx=0, show=False, colors=dict_mapping
            )

    def test_spikes_raster_dipole_overlay(self, base_simulation_spikes):
        net, dpls = base_simulation_spikes

        # Missing dipole argument raises error
        # --------------------------------------------------
        with pytest.raises(ValueError, match="Dipole object must be provided"):
            net.cell_response.plot_spikes_raster(
                overlay_dipoles=True,
                dpl=None,
            )

        # Confirm dipoles are scaled correctly
        # --------------------------------------------------
        # Get initial y-axis ticks for raster plot without dipole
        fig = net.cell_response.plot_spikes_raster()
        initial_raster_yrange = fig.axes[0].get_yticks()

        # Get initial y-axis range, including a small allowance for a single
        # increment expansion on either end of the axis
        increment = abs(initial_raster_yrange[1] - initial_raster_yrange[0])
        initial_raster_yrange = (
            abs(max(initial_raster_yrange) - min(initial_raster_yrange)) + increment * 2
        )

        # Ensure dipoles are initially out of the bounds of the raster plot
        for dipole in dpls:
            for layer in ["L2", "L5"]:
                dipole.data[layer] = dipole.data[layer] * initial_raster_yrange

        # Get y-axis range of raster plot with overlaid dipoles
        fig = net.cell_response.plot_spikes_raster(
            overlay_dipoles=True,
            dpl=dpls,
        )
        updated_raster_yrange = fig.axes[0].get_yticks()
        updated_raster_yrange = abs(
            max(updated_raster_yrange) - min(updated_raster_yrange)
        )

        assert updated_raster_yrange <= initial_raster_yrange


def test_network_plotter_init(setup_net):
    """Test init keywords of NetworkPlotter class."""
    net = setup_net
    # test NetworkPlotter class
    args = [
        "xlim",
        "ylim",
        "zlim",
        "elev",
        "azim",
        "vmin",
        "vmax",
        "trial_idx",
        "time_idx",
        "colorbar",
    ]
    for arg in args:
        with pytest.raises(TypeError, match=f"{arg} must be"):
            net_plot = NetworkPlotter(net, **{arg: "blah"})

    net_plot = NetworkPlotter(net)

    assert net_plot.vsec_array.shape == (159, 1)
    assert net_plot.color_array.shape == (159, 1, 4)
    assert net_plot._vsec_recorded is False
    plt.close("all")


def test_network_plotter_simulation(setup_net):
    """Test NetworkPlotter class simulation warnings."""
    net = setup_net
    net_plot = NetworkPlotter(net)
    # Errors if vsec isn't recorded
    with pytest.raises(RuntimeError, match="Network must be simulated"):
        net_plot.export_movie("demo.gif", dpi=200)

    # Errors if vsec isn't recorded with record_vsec='all'
    _ = simulate_dipole(net, dt=0.5, tstop=10, record_vsec="soma")
    net_plot = NetworkPlotter(net)

    assert net_plot.vsec_array.shape == (159, 1)
    assert net_plot.color_array.shape == (159, 1, 4)
    assert net_plot._vsec_recorded is False

    with pytest.raises(RuntimeError, match="Network must be simulated"):
        net_plot.export_movie("demo.gif", dpi=200)

    net = setup_net
    _ = simulate_dipole(net, dt=0.5, tstop=10, record_vsec="all", n_trials=2)
    net_plot = NetworkPlotter(net)
    # setter/getter test for time_idx and trial_idx
    net_plot.time_idx = 5
    assert net_plot.time_idx == 5
    net_plot.trial_idx = 1
    assert net_plot.trial_idx == 1

    assert net_plot.vsec_array.shape == (159, 21)
    assert net_plot.color_array.shape == (159, 21, 4)
    assert net_plot._vsec_recorded is True
    assert isinstance(net_plot._cbar, Colorbar)
    plt.close("all")


def test_network_plotter_setter(setup_net):
    """Test NetworkPlotter class setters and getters."""
    net = setup_net
    net_plot = NetworkPlotter(net)
    # Type check errors
    args = [
        "xlim",
        "ylim",
        "zlim",
        "elev",
        "azim",
        "vmin",
        "vmax",
        "trial_idx",
        "time_idx",
        "colorbar",
    ]
    for arg in args:
        with pytest.raises(TypeError, match=f"{arg} must be"):
            setattr(net_plot, arg, "blah")

    # Check that the setters and getters work
    arg_dict = {
        "xlim": (-100, 100),
        "ylim": (-100, 100),
        "zlim": (-100, 100),
        "elev": 10,
        "azim": 10,
        "vmin": 0,
        "vmax": 100,
        "bgcolor": "white",
        "voltage_colormap": "jet",
        "colorbar": False,
    }
    for arg, val in arg_dict.items():
        setattr(net_plot, arg, val)
        assert getattr(net_plot, arg) == val

    assert net_plot._cbar is None
    assert net_plot.fig.get_facecolor() == (1.0, 1.0, 1.0, 1.0)

    # time_idx setter should raise an error if network is not simulated
    with pytest.raises(RuntimeError, match="Network must be simulated"):
        net_plot.time_idx = 5

    with pytest.raises(RuntimeError, match="Network must be simulated"):
        net_plot.trial_idx = 1
    plt.close("all")


def test_network_plotter_export(tmp_path, setup_net):
    """Test NetworkPlotter class export methods."""
    net = setup_net
    _ = simulate_dipole(net, dt=0.5, tstop=10, n_trials=1, record_vsec="all")
    net_plot = NetworkPlotter(net)

    # Check no file is already written
    path_out = tmp_path / "demo.gif"
    assert not path_out.is_file()

    # Test animation export and voltage plotting
    net_plot.export_movie(path_out, dpi=200, decim=100, writer="pillow")

    assert path_out.is_file()

    plt.close("all")


def test_invert_spike_types(setup_net):
    """Test plotting a histogram with an inverted external drive"""
    net = setup_net

    weights_ampa = {"L2_pyramidal": 0.15, "L5_pyramidal": 0.15}
    syn_delays = {"L2_pyramidal": 0.1, "L5_pyramidal": 1.0}

    net.add_evoked_drive(
        "evdist1",
        mu=63.53,
        sigma=3.85,
        numspikes=1,
        weights_ampa=weights_ampa,
        location="distal",
        synaptic_delays=syn_delays,
        event_seed=274,
    )

    net.add_evoked_drive(
        "evprox1",
        mu=26.61,
        sigma=2.47,
        numspikes=1,
        weights_ampa=weights_ampa,
        location="proximal",
        synaptic_delays=syn_delays,
        event_seed=274,
    )

    _ = simulate_dipole(net, dt=0.5, tstop=80.0, n_trials=1)

    # test string input
    net.cell_response.plot_spikes_hist(
        spike_types=["evprox", "evdist"],
        invert_spike_types="evdist",
        show=False,
    )

    # test case where all inputs are flipped
    net.cell_response.plot_spikes_hist(
        spike_types=["evprox", "evdist"],
        invert_spike_types=["evprox", "evdist"],
        show=False,
    )

    # test case where some inputs are flipped
    fig = net.cell_response.plot_spikes_hist(
        spike_types=["evprox", "evdist"],
        invert_spike_types=["evdist"],
        show=False,
    )

    # check that there are 2 y axes
    assert len(fig.axes) == 2

    # check for equivalency of both y axes
    y1 = fig.axes[0]
    y2 = fig.axes[1]

    y1_max = max(y1.get_ylim())
    y2_max = max(y2.get_ylim())

    assert y1_max == y2_max

    # check that data are plotted
    assert y1_max > 1

    plt.close("all")



"""
=================================================
05. Optimize simulated evoked response parameters
=================================================

This example demonstrates how to optimize the parameters
of the model simulation to match an experimental dipole waveform.
"""

# Authors: Carolina Fernandez <cxf418@miami.edu>
#          Nick Tolley <nicholas_tolley@brown.edu>
#          Ryan Thorpe <ryan_thorpe@brown.edu>
#          Mainak Jas <mjas@mgh.harvard.edu>

import os.path as op

import matplotlib.pyplot as plt

###############################################################################
# Let us import hnn_core

import hnn_core
from hnn_core import (MPIBackend, jones_2009_model, simulate_dipole,
                      read_dipole)

hnn_core_root = op.join(op.dirname(hnn_core.__file__))

# The number of cores may need modifying depending on your current machine.
n_procs = 10

###############################################################################
# First, we will load experimental data into Dipole object.
#
# This is a different experiment than the one to which the base parameters were
# tuned. So, the initial RMSE will be large, giving the optimization procedure
# a lot to work with.

from urllib.request import urlretrieve

data_url = ('https://raw.githubusercontent.com/jonescompneurolab/hnn/master/'
            'data/MEG_detection_data/S1_SupraT.txt')
urlretrieve(data_url, 'S1_SupraT.txt')
exp_dpl = read_dipole('S1_SupraT.txt')

###############################################################################
# Let’s then simulate the dipole with some initial parameters.

tstop = exp_dpl.times[-1]
scale_factor = 3000
smooth_window_len = 30

net_init = jones_2009_model()

# Proximal 1
weights_ampa_p1 = {'L2_basket': 0.2913, 'L2_pyramidal': 0.9337,
                   'L5_basket': 0.1951, 'L5_pyramidal': 0.3602}
weights_nmda_p1 = {'L2_basket': 0.9240, 'L2_pyramidal': 0.0845,
                   'L5_basket': 0.5849, 'L5_pyramidal': 0.65105}
synaptic_delays_p = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                     'L5_basket': 1., 'L5_pyramidal': 1.}
net_init.add_evoked_drive('evprox1',
                          mu=5.6813,
                          sigma=20.3969,
                          numspikes=1,
                          location='proximal',
                          weights_ampa=weights_ampa_p1,
                          weights_nmda=weights_nmda_p1,
                          synaptic_delays=synaptic_delays_p)

# Distal
weights_ampa_d1 = {'L2_basket': 0.8037, 'L2_pyramidal': 0.5738,
                   'L5_pyramidal': 0.3626}
weights_nmda_d1 = {'L2_basket': 0.2492, 'L2_pyramidal': 0.6183,
                   'L5_pyramidal': 0.1121}
synaptic_delays_d1 = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                      'L5_pyramidal': 0.1}
net_init.add_evoked_drive('evdist1',
                          mu=58.6539,
                          sigma=5.5810,
                          numspikes=1,
                          location='distal',
                          weights_ampa=weights_ampa_d1,
                          weights_nmda=weights_nmda_d1,
                          synaptic_delays=synaptic_delays_d1)

# Proximal 2
weights_ampa_p2 = {'L2_basket': 0.01, 'L2_pyramidal': 0.01, 'L5_basket': 0.01,
                   'L5_pyramidal': 0.01}
weights_nmda_p2 = {'L2_basket': 0.01, 'L2_pyramidal': 0.01, 'L5_basket': 0.01,
                   'L5_pyramidal': 0.01}
net_init.add_evoked_drive('evprox2',
                          mu=80,
                          sigma=1,
                          numspikes=1,
                          location='proximal',
                          weights_ampa=weights_ampa_p2,
                          weights_nmda=weights_nmda_p2,
                          synaptic_delays=synaptic_delays_p)

with MPIBackend(n_procs=n_procs, mpi_cmd='mpiexec'):
    init_dpl = simulate_dipole(net_init, tstop=tstop, n_trials=1)[0]
init_dpl.scale(scale_factor)
init_dpl.smooth(smooth_window_len)

###############################################################################
# Now we start the optimization!
#
# First, we define a function that will tell the optimization routine how to
# modify the network drive parameters. The function will take in the Network
# object with no attached drives, and a dictionary of the parameters we wish to
# optimize.


def set_params(net, params):

    # Proximal 1
    net.add_evoked_drive('evprox1',
                         mu=5.6813,
                         sigma=20.3969,
                         numspikes=1,
                         location='proximal',
                         weights_ampa=weights_ampa_p1,
                         weights_nmda=weights_nmda_p1,
                         synaptic_delays=synaptic_delays_p)

    # Distal
    net.add_evoked_drive('evdist1',
                         mu=58.6539,
                         sigma=5.5810,
                         numspikes=1,
                         location='distal',
                         weights_ampa=weights_ampa_d1,
                         weights_nmda=weights_nmda_d1,
                         synaptic_delays=synaptic_delays_d1)

    # Proximal 2
    weights_ampa_p2 = {'L2_basket':
                       params['evprox2_ampa_L2_basket'],
                       'L2_pyramidal':
                       params['evprox2_ampa_L2_pyramidal'],
                       'L5_basket':
                       params['evprox2_ampa_L5_basket'],
                       'L5_pyramidal':
                       params['evprox2_ampa_L5_pyramidal']}
    weights_nmda_p2 = {'L2_basket':
                       params['evprox2_nmda_L2_basket'],
                       'L2_pyramidal':
                       params['evprox2_nmda_L2_pyramidal'],
                       'L5_basket':
                       params['evprox2_nmda_L5_basket'],
                       'L5_pyramidal':
                       params['evprox2_nmda_L5_pyramidal']}
    net.add_evoked_drive('evprox2',
                         mu=params['evprox2_mu'],
                         sigma=params['evprox2_sigma'],
                         numspikes=1,
                         location='proximal',
                         weights_ampa=weights_ampa_p2,
                         weights_nmda=weights_nmda_p2,
                         synaptic_delays=synaptic_delays_p)

###############################################################################
# Then, we define the constraints.
#
# The constraints must be a dictionary of tuples where the first value in each
# tuple is the lower bound and the second value is the upper bound for the
# corresponding parameter.
#
# The following synaptic weight parameter ranges (units of micro-siemens)
# were chosen so as to keep the model in physiologically realistic regimes.


constraints = dict({'evprox2_ampa_L2_basket': (0.01, 1.),
                    'evprox2_ampa_L2_pyramidal': (0.01, 1.),
                    'evprox2_ampa_L5_basket': (0.01, 1.),
                    'evprox2_ampa_L5_pyramidal': (0.01, 1.),
                    'evprox2_nmda_L2_basket': (0.01, 1.),
                    'evprox2_nmda_L2_pyramidal': (0.01, 1.),
                    'evprox2_nmda_L5_basket': (0.01, 1.),
                    'evprox2_nmda_L5_pyramidal': (0.01, 1.),
                    'evprox2_mu': (100., 120.),
                    'evprox2_sigma': (2., 30.)})

###############################################################################
# Now we define and fit the optimizer.

from hnn_core.optimization import Optimizer

net = jones_2009_model()
optim = Optimizer(net, tstop=tstop, constraints=constraints,
                  set_params=set_params)
with MPIBackend(n_procs=n_procs, mpi_cmd='mpiexec'):
    optim.fit(target=exp_dpl, scale_factor=scale_factor,
              smooth_window_len=smooth_window_len)

###############################################################################
# Finally, we can plot the experimental data alongside the post-optimization
# simulation dipole as well as the convergence plot.

with MPIBackend(n_procs=n_procs, mpi_cmd='mpiexec'):
    opt_dpl = simulate_dipole(optim.net_, tstop=tstop, n_trials=1)[0]
opt_dpl.scale(scale_factor)
opt_dpl.smooth(smooth_window_len)

fig, axes = plt.subplots(2, 1, sharex=True, figsize=(6, 6))

# plot original
exp_dpl.plot(ax=axes[0], layer='agg', show=False, color='tab:blue')
init_dpl.plot(ax=axes[0], layer='agg', show=False, color='tab:orange')
opt_dpl.plot(ax=axes[0], layer='agg', show=False, color='tab:green')
axes[0].legend(['experimental', 'initial', 'optimized'])
optim.net_.cell_response.plot_spikes_hist(ax=axes[1])

fig1 = optim.plot_convergence()



"""Network io"""
# Authors: Rajat Partani <rajatpartani@gmail.com>
#          Mainak Jas <mjas@mgh.harvard.edu>
#          Nick Tolley <nicholas_tolley@brown.edu>
#          George Dang <george_dang@brown.edu>

import os
import json
import numpy as np

from collections import OrderedDict
from pathlib import Path

from .cell import Cell, Section
from .cell_response import CellResponse
from .externals.mne import fill_doc


def _convert_np_array_to_list(obj):
    """Returns object with np.arrays converted to lists

    Converts np.arrays to lists. Dicts or lists with nested np.arrays will
    have nested arrays converted to lists.
    """
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, dict):
        return {k: _convert_np_array_to_list(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_convert_np_array_to_list(item) for item in obj]
    else:
        return obj


def _cell_response_to_dict(net, write_output):
    """Returns a dict of cell response data."""
    # Write cell_response as dict
    if (not net.cell_response) or (not write_output):
        return dict()
    else:
        return net.cell_response.to_dict()


def _rec_array_to_dict(value, write_output):
    """Returns a dict of rec_array data."""
    rec_array_copy = value.copy()
    if not write_output:
        rec_array_copy._reset()
    rec_array_copy_dict = rec_array_copy.to_dict()
    return rec_array_copy_dict


def _conn_to_dict(conn):
    """Converts a Connectivity object parameters to a dict format."""
    conn_data = {
        "target_type": conn["target_type"],
        "target_gids": list(conn["target_gids"]),
        "num_targets": conn["num_targets"],
        "src_type": conn["src_type"],
        "src_gids": list(conn["src_gids"]),
        "num_srcs": conn["num_srcs"],
        "gid_pairs": {str(key): val for key, val in conn["gid_pairs"].items()},
        "loc": conn["loc"],
        "receptor": conn["receptor"],
        "nc_dict": conn["nc_dict"],
        "allow_autapses": int(conn["allow_autapses"]),
        "probability": conn["probability"],
    }
    return conn_data


def _external_drive_to_dict(drive, write_output):
    """Returns dict of drive data from a Drive object."""
    drive_data = dict()
    for key in drive.keys():
        # Cannot store sets with hdf5
        if isinstance(drive[key], set):
            drive_data[key] = list(drive[key])
        else:
            drive_data[key] = drive[key]
    if not write_output:
        drive_data["events"] = list()
    return drive_data


def _str_to_node(node_string):
    """Returns tuple of node values from a comma-separated string format."""
    node_tuple = node_string.split(",")
    node_tuple[1] = int(node_tuple[1])
    node = (node_tuple[0], node_tuple[1])
    return node


def _read_cell_types(cell_types_data):
    """Returns a dict of Cell objects from json encoded data

    This function handles both legacy format (direct cell data) and
    new format (cell object with metadata) for backwards compatibility.

    Parameters
    ----------
    cell_types_data : dict
        Dictionary containing cell type information in either:
        - Legacy format: {cell_name: cell_dict} where cell_dict contains
          cell properties directly
        - New format: {cell_name: {"cell_object": cell_dict, "cell_metadata": metadata}}
          where cell_dict contains cell properties and metadata contains additional info

    Returns
    -------
    cell_types : dict
        Dictionary with cell names as keys and dicts containing:
        - "cell_object": Cell instance
        - "cell_metadata": dict of metadata (empty dict for legacy format)
    """
    cell_types = dict()
    for cell_name in cell_types_data:
        # Determine format and extract cell_data and metadata accordingly
        if (
            "cell_object" in cell_types_data[cell_name]
            and "cell_metadata" in cell_types_data[cell_name]
        ):
            # Format post-commit-6388f9f:
            #   Extract cell properties from nested "cell_object"
            cell_data = cell_types_data[cell_name]["cell_object"]
            cell_metadata = cell_types_data[cell_name]["cell_metadata"]
        else:
            # Format pre-commit-6388f9f:
            #   Treat the entire cell_data as the cell information
            cell_data = cell_types_data[cell_name]
            if cell_name == "L2_basket":
                cell_metadata = {
                    "morpho_type": "basket",
                    "electro_type": "inhibitory",
                    "layer": "2",
                    "measure_dipole": False,
                    "reference": "https://doi.org/10.7554/eLife.51214",
                }
            elif cell_name == "L2_pyramidal":
                cell_metadata = {
                    "morpho_type": "pyramidal",
                    "electro_type": "excitatory",
                    "layer": "2",
                    "measure_dipole": True,
                    "reference": "https://doi.org/10.7554/eLife.51214",
                }
            elif cell_name == "L5_basket":
                cell_metadata = {
                    "morpho_type": "basket",
                    "electro_type": "inhibitory",
                    "layer": "5",
                    "measure_dipole": False,
                    "reference": "https://doi.org/10.7554/eLife.51214",
                }
            elif cell_name == "L5_pyramidal":
                cell_metadata = {
                    "morpho_type": "pyramidal",
                    "electro_type": "excitatory",
                    "layer": "5",
                    "measure_dipole": True,
                    "reference": "https://doi.org/10.7554/eLife.51214",
                }

        # Now cell_data contains the cell properties regardless of format
        sections = dict()
        sections_data = cell_data["sections"]
        for section_name in sections_data:
            section_data = sections_data[section_name]
            sections[section_name] = Section(
                L=section_data["L"],
                diam=section_data["diam"],
                cm=section_data["cm"],
                Ra=section_data["Ra"],
                end_pts=section_data["end_pts"],
            )
            # Set section attributes
            sections[section_name].syns = section_data["syns"]
            sections[section_name].mechs = section_data["mechs"]

        # cell_tree
        cell_tree = None
        if cell_data["cell_tree"] is not None:
            cell_tree = dict()
            for parent, children in cell_data["cell_tree"].items():
                key = _str_to_node(parent)
                value = list()
                for child in children:
                    value.append(_str_to_node(child))
                cell_tree[key] = value

        cell_object = Cell(
            name=cell_data["name"],
            pos=tuple(cell_data["pos"]),
            sections=sections,
            synapses=cell_data["synapses"],
            cell_tree=cell_tree,
            sect_loc=cell_data["sect_loc"],
            gid=cell_data["gid"],
        )

        # Set additional cell attributes
        cell_object.dipole_pp = cell_data["dipole_pp"]
        cell_object.vsec = cell_data["vsec"]
        cell_object.isec = cell_data["isec"]
        cell_object.tonic_biases = cell_data["tonic_biases"]

        # Store in the unified format with cell_metadata
        cell_types[cell_name] = {
            "cell_object": cell_object,
            "cell_metadata": cell_metadata,
        }

    return cell_types


def _read_cell_response(cell_response_data, read_output):
    """Returns CellResponse from json encoded data"""
    if (not cell_response_data) or (not read_output):
        return None
    cell_response = CellResponse(
        cell_type_names=cell_response_data["cell_type_names"],
        spike_times=cell_response_data["spike_times"],
        spike_gids=cell_response_data["spike_gids"],
        spike_types=cell_response_data["spike_types"],
    )

    cell_response._times = cell_response_data["times"]
    cell_response._vsec = list()
    for trial in cell_response_data["vsec"]:
        trial = dict((int(key), val) for key, val in trial.items())
        cell_response._vsec.append(trial)
    cell_response._isec = list()
    for trial in cell_response_data["isec"]:
        trial = dict((int(key), val) for key, val in trial.items())
        cell_response._isec.append(trial)
    return cell_response


def _set_from_cell_specific(drive_data):
    """Returns number of drive cells based on cell_specific bool

    The n_drive_cells keyword for add_poisson_drive and add_bursty_drive
    methods accept either an int or string (n_cells). If the bool keyword
    cell_specific = True, n_drive_cells must be 'n_cells'.
    """
    if drive_data["cell_specific"]:
        return "n_cells"
    return drive_data["n_drive_cells"]


def _read_external_drive(net, drive_data, read_output):
    """Adds drives encoded in json data to a Network"""

    if (drive_data["type"] == "evoked") or (drive_data["type"] == "gaussian"):
        # Skipped n_drive_cells here
        net.add_evoked_drive(
            name=drive_data["name"],
            mu=drive_data["dynamics"]["mu"],
            sigma=drive_data["dynamics"]["sigma"],
            numspikes=drive_data["dynamics"]["numspikes"],
            location=drive_data["location"],
            n_drive_cells=_set_from_cell_specific(drive_data),
            cell_specific=drive_data["cell_specific"],
            weights_ampa=drive_data["weights_ampa"],
            weights_nmda=drive_data["weights_nmda"],
            synaptic_delays=drive_data["synaptic_delays"],
            probability=drive_data["probability"],
            event_seed=drive_data["event_seed"],
            conn_seed=drive_data["conn_seed"],
        )
    elif drive_data["type"] == "poisson":
        net.add_poisson_drive(
            name=drive_data["name"],
            tstart=drive_data["dynamics"]["tstart"],
            tstop=drive_data["dynamics"]["tstop"],
            rate_constant=(drive_data["dynamics"]["rate_constant"]),
            location=drive_data["location"],
            n_drive_cells=(_set_from_cell_specific(drive_data)),
            cell_specific=drive_data["cell_specific"],
            weights_ampa=drive_data["weights_ampa"],
            weights_nmda=drive_data["weights_nmda"],
            synaptic_delays=drive_data["synaptic_delays"],
            probability=drive_data["probability"],
            event_seed=drive_data["event_seed"],
            conn_seed=drive_data["conn_seed"],
        )
    elif drive_data["type"] == "bursty":
        net.add_bursty_drive(
            name=drive_data["name"],
            tstart=drive_data["dynamics"]["tstart"],
            tstart_std=drive_data["dynamics"]["tstart_std"],
            tstop=drive_data["dynamics"]["tstop"],
            burst_rate=drive_data["dynamics"]["burst_rate"],
            burst_std=drive_data["dynamics"]["burst_std"],
            numspikes=drive_data["dynamics"]["numspikes"],
            spike_isi=drive_data["dynamics"]["spike_isi"],
            location=drive_data["location"],
            n_drive_cells=_set_from_cell_specific(drive_data),
            cell_specific=drive_data["cell_specific"],
            weights_ampa=drive_data["weights_ampa"],
            weights_nmda=drive_data["weights_nmda"],
            synaptic_delays=drive_data["synaptic_delays"],
            probability=drive_data["probability"],
            event_seed=drive_data["event_seed"],
            conn_seed=drive_data["conn_seed"],
        )

    net.external_drives[drive_data["name"]]["events"] = drive_data["events"]
    if not read_output:
        net.external_drives[drive_data["name"]]["events"] = list()


def _read_connectivity(net, conns_data):
    """Adds connections to a Network from json encoded connectivity"""
    # Overwrite drive connections
    net.connectivity = list()

    for conn_data in conns_data:
        src_gids = [int(s) for s in conn_data["gid_pairs"].keys()]
        target_gids_nested = [
            target_gid for target_gid in conn_data["gid_pairs"].values()
        ]
        conn_data["allow_autapses"] = bool(conn_data["allow_autapses"])
        net.add_connection(
            src_gids=src_gids,
            target_gids=target_gids_nested,
            loc=conn_data["loc"],
            receptor=conn_data["receptor"],
            weight=conn_data["nc_dict"]["A_weight"],
            delay=conn_data["nc_dict"]["A_delay"],
            lamtha=conn_data["nc_dict"]["lamtha"],
            allow_autapses=conn_data["allow_autapses"],
            probability=conn_data["probability"],
        )


def _read_rec_arrays(net, rec_arrays_data, read_output):
    """Adds rec arrays to Network from json data."""
    for key in rec_arrays_data:
        rec_array = rec_arrays_data[key]
        net.add_electrode_array(
            name=key,
            electrode_pos=[tuple(pos) for pos in rec_array["positions"]],
            conductivity=rec_array["conductivity"],
            method=rec_array["method"],
            min_distance=rec_array["min_distance"],
        )
        net.rec_arrays[key]._times = rec_array["times"]
        net.rec_arrays[key]._data = rec_array["voltages"]
        if not read_output:
            net.rec_arrays[key]._reset()


def _read_pos_dict(pos_dict):
    """Returns position dictionary with nested positions converted to tuple."""
    pos_dict_converted = dict()
    for key, value in pos_dict.items():
        if key == "origin":
            pos_dict_converted[key] = tuple(value)
        else:
            pos_dict_converted[key] = [tuple(position) for position in value]
    return pos_dict_converted


def network_to_dict(net, write_output=False):
    """Returns a dict of parameters and outputs from Network.

    Parameters
    ----------
    net : Network
        hnn-core Network object
    write_output : bool
        Includes simulation-associated data.
    Returns
    -------
    dict
    """

    # cell_types serialization, support both old and new formats
    cell_types_data = {}
    for name, template in net.cell_types.items():
        if isinstance(template, dict) and "cell_object" in template:
            # New format with cell_metadata
            cell_types_data[name] = {
                "cell_object": template["cell_object"].to_dict(),
                "cell_metadata": template["cell_metadata"],
            }
        else:
            # Legacy format, template is a Cell object directly
            cell_types_data[name] = template.to_dict()

    net_data = {
        "object_type": "Network",
        "legacy_mode": net._legacy_mode,
        "N_pyr_x": net._N_pyr_x,
        "N_pyr_y": net._N_pyr_y,
        "celsius": net._params["celsius"],
        "cell_types": cell_types_data,
        "gid_ranges": {
            cell: {"start": c_range.start, "stop": c_range.stop}
            for cell, c_range in net.gid_ranges.items()
        },
        "pos_dict": {cell: pos for cell, pos in net.pos_dict.items()},
        "cell_response": _cell_response_to_dict(net, write_output),
        "external_drives": {
            drive: _external_drive_to_dict(params, write_output)
            for drive, params in net.external_drives.items()
        },
        "external_biases": net.external_biases,
        "connectivity": [_conn_to_dict(conn) for conn in net.connectivity],
        "rec_arrays": {
            ra_name: _rec_array_to_dict(ex_array, write_output)
            for ra_name, ex_array in net.rec_arrays.items()
        },
        "threshold": net.threshold,
        "delay": net.delay,
    }
    return net_data


@fill_doc
def write_network_configuration(net, output, overwrite=True):
    """Writes network configuration to a json file.

    Writes network configurations as a hierarchical json similar to the Network
    object's structure. Outputs recorded during simulation such as currents and
    voltages are not saved due to size.

    Parameters
    ----------
    net : Network
        hnn-core Network object
    output : str, Path, or StringIO
        Path or buffer to write outputs
    overwrite : bool
        Overwrite file if it exists. Default: True

    Returns
    -------
    None
    """

    net_data = net.to_dict(write_output=False)
    net_data_converted = _convert_np_array_to_list(net_data)

    if isinstance(output, (str, Path)):
        if overwrite is False and os.path.exists(output):
            raise FileExistsError(
                "File already exists at path %s. Rename "
                "the file or set overwrite=True." % (output,)
            )
        # Saving file
        with open(output, "w", encoding="utf-8") as f:
            json.dump(net_data_converted, f, ensure_ascii=False, indent=4)
    else:
        # Write to StringIO buffer
        json.dump(net_data_converted, output, ensure_ascii=False, indent=4)
        output.seek(0)  # Reset buffer position to the start


def _order_drives(gid_ranges, external_drives):
    """Returns an ordered dict of external drives by ascending gid ranges

    Drive data from hdf5 are ordered alphabetically by name. This function
    reorders the external drives by ascending gid ranges.

    Parameters
    ----------
    gid_ranges : dict (keys: names) of range
        Dictionary with connection or drive names as keys and ranges as values.
    external_drives: dict (keys: drive names) of dict (keys: parameters)
        Dictionary with drive name as keys and instances of _NetworkDrive as
        values.

    Returns
    -------
    OrderedDict : dict (keys: drive names) of dict (keys: parameters)
        Ordered dict with drives by ascending gid ranges
    """
    ordered_drives = OrderedDict()
    min_gid_to_drive = {
        min(gid_range): name
        for (name, gid_range) in gid_ranges.items()
        if name in external_drives.keys()
    }
    min_gid_sorted = sorted(list(min_gid_to_drive.keys()))
    for min_gid in min_gid_sorted:
        drive_name = min_gid_to_drive[min_gid]
        ordered_drives[drive_name] = external_drives[drive_name]

    return ordered_drives


def dict_to_network(net_data, read_drives=True, read_external_biases=True):
    """Converts a dict of network configurations to a Network

    Parameters
    ----------
    net_data : dict
        Dictionary containing network configurations.
    read_drives : bool, optional
        Read-in drives to Network object. Default is True.
    read_external_biases : bool, optional
        Read-in external biases to Network object. Default is True.

    Returns : Network
    -------
    """

    # Importing Network.
    # Cannot do this globally due to circular import.
    from .network import Network

    params = dict()
    params["celsius"] = net_data["celsius"]
    params["threshold"] = net_data["threshold"]

    mesh_shape = (net_data["N_pyr_x"], net_data["N_pyr_y"])

    # Instantiating network
    net = Network(params, mesh_shape=mesh_shape, legacy_mode=net_data["legacy_mode"])

    # Setting attributes
    # Set cell types
    net.cell_types = _read_cell_types(net_data["cell_types"])
    # Set gid ranges
    gid_ranges_data = dict()
    for key in net_data["gid_ranges"]:
        start = net_data["gid_ranges"][key]["start"]
        stop = net_data["gid_ranges"][key]["stop"]
        gid_ranges_data[key] = range(start, stop)
    net.gid_ranges = OrderedDict(gid_ranges_data)
    # Set pos_dict
    net.pos_dict = _read_pos_dict(net_data["pos_dict"])
    # Set cell_response
    net.cell_response = _read_cell_response(
        net_data["cell_response"], read_output=False
    )
    # Set external drives
    external_drive_data = _order_drives(net.gid_ranges, net_data["external_drives"])
    for key in external_drive_data.keys():
        _read_external_drive(net, external_drive_data[key], read_output=False)
    # Set external biases
    if read_external_biases:
        net.external_biases = net_data["external_biases"]
    # Set connectivity
    _read_connectivity(net, net_data["connectivity"])
    # Set rec_arrays
    _read_rec_arrays(net, net_data["rec_arrays"], read_output=False)
    # Set threshold
    net.threshold = net_data["threshold"]
    # Set delay
    net.delay = net_data["delay"]

    if not read_drives:
        net.clear_drives()

    return net


def read_network_configuration(fname, read_drives=True, read_external_biases=True):
    """Read network from a json configuration file.

    Parameters
    ----------
    fname : str or Path
        Path to configuration file
    read_drives : bool
        Read-in drives to Network object
    read_external_biases
        Read-in external biases to Network object

    Returns : Network
    -------

    """

    with open(fname, "r") as file:
        net_data = json.load(file)

    if net_data.get("object_type") != "Network":
        raise ValueError(
            "The json should encode a Network object. "
            "The file contains object of "
            "type %s" % (net_data.get("object_type"))
        )

    net = dict_to_network(net_data, read_drives, read_external_biases)

    return net



"""Master list of changeable params."""

# Authors: Mainak Jas <mjas@mgh.harvard.edu>
#          Sam Neymotin <samnemo@gmail.com>


def get_params_default(nprox=2, ndist=1):
    """Note that nearly all start times are set BEYOND tstop for this file
    Most values here are set to whatever default value
    inactivates them, such as 0 for conductance
    prng seed values are also set to 0 (non-random)
    flat file of default values
    will most often be overwritten
    """
    # set default params
    p = {
        "sim_prefix": "default",
        # simulation end time (ms)
        "tstop": 250.0,
        # numbers of cells making up the pyramidal grids
        "N_pyr_x": 1,
        "N_pyr_y": 1,
        # amplitudes of individual Gaussian random inputs to L2Pyr and L5Pyr
        # L2 Basket params
        "L2Basket_Gauss_A_weight": 0.0,
        "L2Basket_Gauss_mu": 2000.0,
        "L2Basket_Gauss_sigma": 3.6,
        "L2Basket_Pois_A_weight_ampa": 0.0,
        "L2Basket_Pois_A_weight_nmda": 0.0,
        "L2Basket_Pois_lamtha": 0.0,
        # L2 Pyr params
        "L2Pyr_Gauss_A_weight": 0.0,
        "L2Pyr_Gauss_mu": 2000.0,
        "L2Pyr_Gauss_sigma": 3.6,
        "L2Pyr_Pois_A_weight_ampa": 0.0,
        "L2Pyr_Pois_A_weight_nmda": 0.0,
        "L2Pyr_Pois_lamtha": 0.0,
        # L5 Pyr params
        "L5Pyr_Gauss_A_weight": 0.0,
        "L5Pyr_Gauss_mu": 2000.0,
        "L5Pyr_Gauss_sigma": 4.8,
        "L5Pyr_Pois_A_weight_ampa": 0.0,
        "L5Pyr_Pois_A_weight_nmda": 0.0,
        "L5Pyr_Pois_lamtha": 0.0,
        # L5 Basket params
        "L5Basket_Gauss_A_weight": 0.0,
        "L5Basket_Gauss_mu": 2000.0,
        "L5Basket_Gauss_sigma": 2.0,
        "L5Basket_Pois_A_weight_ampa": 0.0,
        "L5Basket_Pois_A_weight_nmda": 0.0,
        "L5Basket_Pois_lamtha": 0.0,
        # maximal conductances for all synapses
        # max conductances TO L2Pyrs
        "gbar_L2Pyr_L2Pyr_ampa": 0.0,
        "gbar_L2Pyr_L2Pyr_nmda": 0.0,
        "gbar_L2Basket_L2Pyr_gabaa": 0.0,
        "gbar_L2Basket_L2Pyr_gabab": 0.0,
        # max conductances TO L2Baskets
        "gbar_L2Pyr_L2Basket": 0.0,
        "gbar_L2Basket_L2Basket": 0.0,
        # max conductances TO L5Pyr
        "gbar_L5Pyr_L5Pyr_ampa": 0.0,
        "gbar_L5Pyr_L5Pyr_nmda": 0.0,
        "gbar_L2Pyr_L5Pyr": 0.0,
        "gbar_L2Basket_L5Pyr": 0.0,
        "gbar_L5Basket_L5Pyr_gabaa": 0.0,
        "gbar_L5Basket_L5Pyr_gabab": 0.0,
        # max conductances TO L5Baskets
        "gbar_L5Basket_L5Basket": 0.0,
        "gbar_L5Pyr_L5Basket": 0.0,
        "gbar_L2Pyr_L5Basket": 0.0,
        # Ongoing proximal alpha rhythm
        "distribution_prox": "normal",
        "t0_input_prox": 1000.0,
        "tstop_input_prox": 250.0,
        "f_input_prox": 10.0,
        "f_stdev_prox": 20.0,
        "events_per_cycle_prox": 2,
        "repeats_prox": 10,
        "t0_input_stdev_prox": 0.0,
        # Ongoing distal alpha rhythm
        "distribution_dist": "normal",
        "t0_input_dist": 1000.0,
        "tstop_input_dist": 250.0,
        "f_input_dist": 10.0,
        "f_stdev_dist": 20.0,
        "events_per_cycle_dist": 2,
        "repeats_dist": 10,
        "t0_input_stdev_dist": 0.0,
        # thalamic input amplitudes and delays
        "input_prox_A_weight_L2Pyr_ampa": 0.0,
        "input_prox_A_weight_L2Pyr_nmda": 0.0,
        "input_prox_A_weight_L5Pyr_ampa": 0.0,
        "input_prox_A_weight_L5Pyr_nmda": 0.0,
        "input_prox_A_weight_L2Basket_ampa": 0.0,
        "input_prox_A_weight_L2Basket_nmda": 0.0,
        "input_prox_A_weight_L5Basket_ampa": 0.0,
        "input_prox_A_weight_L5Basket_nmda": 0.0,
        "input_prox_A_delay_L2": 0.1,
        "input_prox_A_delay_L5": 1.0,
        # current values, not sure where these distal values come from, need to
        # check
        "input_dist_A_weight_L2Pyr_ampa": 0.0,
        "input_dist_A_weight_L2Pyr_nmda": 0.0,
        "input_dist_A_weight_L5Pyr_ampa": 0.0,
        "input_dist_A_weight_L5Pyr_nmda": 0.0,
        "input_dist_A_weight_L2Basket_ampa": 0.0,
        "input_dist_A_weight_L2Basket_nmda": 0.0,
        "input_dist_A_delay_L2": 5.0,
        "input_dist_A_delay_L5": 5.0,
        # times and stdevs for evoked responses
        "dt_evprox0_evdist": -1,  # not used in GUI
        "dt_evprox0_evprox1": -1,  # not used in GUI
        # whether evoked inputs arrive at same time to all cells
        "sync_evinput": 1,
        # increment (ms) for avg evoked input start (for trial n, avg start
        # time is n * evinputinc
        "inc_evinput": 0.0,
        # analysis
        "save_spec_data": 0,
        "f_max_spec": 40.0,
        "spec_cmap": "jet",  # only used in GUI
        "dipole_scalefctr": 30e3,  # scale factor for dipole - default at 30e3
        # based on scaling needed to match model ongoing rhythms from
        # jones 2009 - for ERPs can use 300
        # for ongoing rhythms + ERPs ... use ... ?
        # window for smoothing (box filter) - 15 ms from jones 2009; shorten
        "dipole_smooth_win": 15.0,
        # in case want to look at higher frequency activity
        "save_figs": 0,
        "save_dpl": 0,  # whether to write dipole output to a file
        "record_vsoma": 0,  # whether to record somatic voltages
        "record_isoma": 0,  # whether to record somatic currents
        "record_vsec": 0,  # whether to record voltages
        "record_isec": 0,  # whether to record currents
        "record_ca": 0,  # whether to record calcium concentration
        # numerics
        # N_trials of 1 means that seed is set by rank
        "N_trials": 1,
        # prng_state is a string for a filename containing the
        # random state one wants to use
        # prng seed cores are the base integer seed for the specific
        # prng object for a specific random number stream
        # 'prng_state': None,
        "prng_seedcore_input_prox": 0,
        "prng_seedcore_input_dist": 0,
        "prng_seedcore_extpois": 0,
        "prng_seedcore_extgauss": 0,
        # default end time for pois inputs
        "t0_pois": 0.0,
        "T_pois": -1,
        "dt": 0.025,
        "celsius": 37.0,
        "threshold": 0.0,  # firing threshold
    }

    # grab cell-specific params and update p accordingly
    p_L2Pyr = get_L2Pyr_params_default()
    p_L5Pyr = get_L5Pyr_params_default()
    p.update(p_L2Pyr)
    p.update(p_L5Pyr)

    # get evoked params and update p accordingly
    p_ev_prox = get_ev_params_default(nprox, True)
    p_ev_dist = get_ev_params_default(ndist, False)
    p.update(p_ev_prox)
    p.update(p_ev_dist)

    return p


# return dict with default params (empty) for evoked inputs;
# n is number of evoked inputs
# isprox == True iff proximal (otherwise distal)


def get_ev_params_default(n, isprox):
    dout = {}  # OrderedDict()
    if isprox:
        pref = "evprox"
    else:
        pref = "evdist"
    lty = ["L2Pyr", "L5Pyr", "L2Basket"]
    if isprox:
        lty.append("L5Basket")
    lsy = ["ampa", "nmda"]  # allow changing both ampa and nmda weights
    for i in range(n):
        tystr = pref + "_" + str(i + 1)  # this string includes input number
        for ty in lty:
            for sy in lsy:
                dout["gbar_" + tystr + "_" + ty + "_" + sy] = 0.0  # feed strength
        dout["t_" + tystr] = 0.0  # times and stdevs for evoked responses
        dout["sigma_t_" + tystr] = 0.0
        # random number generator seed for this input
        dout["prng_seedcore_" + tystr] = 0
        # number of presynaptic spikes (postsynaptic inputs)
        dout["numspikes_" + tystr] = 1
    return dout


def get_L2Pyr_params_default():
    """Returns default params for L2 pyramidal cell."""
    return {
        # Soma
        "L2Pyr_soma_L": 22.1,
        "L2Pyr_soma_diam": 23.4,
        "L2Pyr_soma_cm": 0.6195,
        "L2Pyr_soma_Ra": 200.0,
        # Dendrites
        "L2Pyr_dend_cm": 0.6195,
        "L2Pyr_dend_Ra": 200.0,
        "L2Pyr_apicaltrunk_L": 59.5,
        "L2Pyr_apicaltrunk_diam": 4.25,
        "L2Pyr_apical1_L": 306.0,
        "L2Pyr_apical1_diam": 4.08,
        "L2Pyr_apicaltuft_L": 238.0,
        "L2Pyr_apicaltuft_diam": 3.4,
        "L2Pyr_apicaloblique_L": 340.0,
        "L2Pyr_apicaloblique_diam": 3.91,
        "L2Pyr_basal1_L": 85.0,
        "L2Pyr_basal1_diam": 4.25,
        "L2Pyr_basal2_L": 255.0,
        "L2Pyr_basal2_diam": 2.72,
        "L2Pyr_basal3_L": 255.0,
        "L2Pyr_basal3_diam": 2.72,
        # Synapses
        "L2Pyr_ampa_e": 0.0,
        "L2Pyr_ampa_tau1": 0.5,
        "L2Pyr_ampa_tau2": 5.0,
        "L2Pyr_nmda_e": 0.0,
        "L2Pyr_nmda_tau1": 1.0,
        "L2Pyr_nmda_tau2": 20.0,
        "L2Pyr_gabaa_e": -80.0,
        "L2Pyr_gabaa_tau1": 0.5,
        "L2Pyr_gabaa_tau2": 5.0,
        "L2Pyr_gabab_e": -80.0,
        "L2Pyr_gabab_tau1": 1.0,
        "L2Pyr_gabab_tau2": 20.0,
        # Biophysics soma
        "L2Pyr_soma_gkbar_hh2": 0.01,
        "L2Pyr_soma_gnabar_hh2": 0.18,
        "L2Pyr_soma_el_hh2": -65.0,
        "L2Pyr_soma_gl_hh2": 4.26e-5,
        "L2Pyr_soma_gbar_km": 250.0,
        # Biophysics dends
        "L2Pyr_dend_gkbar_hh2": 0.01,
        "L2Pyr_dend_gnabar_hh2": 0.15,
        "L2Pyr_dend_el_hh2": -65.0,
        "L2Pyr_dend_gl_hh2": 4.26e-5,
        "L2Pyr_dend_gbar_km": 250.0,
    }


def get_L5Pyr_params_default():
    """Returns default params for L5 pyramidal cell."""
    return {
        # Soma
        "L5Pyr_soma_L": 39.0,
        "L5Pyr_soma_diam": 28.9,
        "L5Pyr_soma_cm": 0.85,
        "L5Pyr_soma_Ra": 200.0,
        # Dendrites
        "L5Pyr_dend_cm": 0.85,
        "L5Pyr_dend_Ra": 200.0,
        "L5Pyr_apicaltrunk_L": 102.0,
        "L5Pyr_apicaltrunk_diam": 10.2,
        "L5Pyr_apical1_L": 680.0,
        "L5Pyr_apical1_diam": 7.48,
        "L5Pyr_apical2_L": 680.0,
        "L5Pyr_apical2_diam": 4.93,
        "L5Pyr_apicaltuft_L": 425.0,
        "L5Pyr_apicaltuft_diam": 3.4,
        "L5Pyr_apicaloblique_L": 255.0,
        "L5Pyr_apicaloblique_diam": 5.1,
        "L5Pyr_basal1_L": 85.0,
        "L5Pyr_basal1_diam": 6.8,
        "L5Pyr_basal2_L": 255.0,
        "L5Pyr_basal2_diam": 8.5,
        "L5Pyr_basal3_L": 255.0,
        "L5Pyr_basal3_diam": 8.5,
        # Synapses
        "L5Pyr_ampa_e": 0.0,
        "L5Pyr_ampa_tau1": 0.5,
        "L5Pyr_ampa_tau2": 5.0,
        "L5Pyr_nmda_e": 0.0,
        "L5Pyr_nmda_tau1": 1.0,
        "L5Pyr_nmda_tau2": 20.0,
        "L5Pyr_gabaa_e": -80.0,
        "L5Pyr_gabaa_tau1": 0.5,
        "L5Pyr_gabaa_tau2": 5.0,
        "L5Pyr_gabab_e": -80.0,
        "L5Pyr_gabab_tau1": 1.0,
        "L5Pyr_gabab_tau2": 20.0,
        # Biophysics soma
        "L5Pyr_soma_gkbar_hh2": 0.01,
        "L5Pyr_soma_gnabar_hh2": 0.16,
        "L5Pyr_soma_el_hh2": -65.0,
        "L5Pyr_soma_gl_hh2": 4.26e-5,
        "L5Pyr_soma_gbar_ca": 60.0,
        "L5Pyr_soma_taur_cad": 20.0,
        "L5Pyr_soma_gbar_kca": 2e-4,
        "L5Pyr_soma_gbar_km": 200.0,
        "L5Pyr_soma_gbar_cat": 2e-4,
        "L5Pyr_soma_gbar_ar": 1e-6,
        # Biophysics dends
        "L5Pyr_dend_gkbar_hh2": 0.01,
        "L5Pyr_dend_gnabar_hh2": 0.14,
        "L5Pyr_dend_el_hh2": -71.0,
        "L5Pyr_dend_gl_hh2": 4.26e-5,
        "L5Pyr_dend_gbar_ca": 60.0,
        "L5Pyr_dend_taur_cad": 20.0,
        "L5Pyr_dend_gbar_kca": 2e-4,
        "L5Pyr_dend_gbar_km": 200.0,
        "L5Pyr_dend_gbar_cat": 2e-4,
        "L5Pyr_dend_gbar_ar": 1e-6,
    }



"""Network model functions."""

# Authors: Nick Tolley <nicholas_tolley@brown.edu>

import os.path as op
import hnn_core
from hnn_core import read_params
from .network import Network, _create_cell_coords
from .params import _short_name
from .cells_default import pyramidal_ca, pyramidal, basket
from .externals.mne import _validate_type

# ToDO -> direct _cell_L2Pyr calling


def jones_2009_model(
    params=None,
    add_drives_from_params=False,
    legacy_mode=False,
    mesh_shape=(10, 10),
):
    """Instantiate the network model described in Jones et al. 2009

    Parameters
    ----------
    params : str | dict | None
        The path to the parameter file for constructing the network.
        If None, parameters loaded from default.json
        Default: None
    add_drives_from_params : bool
        If True, add drives as defined in the params-dict. NB this is mainly
        for backward-compatibility with HNN GUI, and will be deprecated in a
        future release. Default: False
    legacy_mode : bool
        Set to False by default. Enables matching HNN GUI output when drives
        are added suitably. Will be deprecated in a future release.
    mesh_shape : tuple of int (default: (10, 10))
        Defines the (n_x, n_y) shape of the grid of pyramidal cells.

    Returns
    -------
    net : Instance of Network object
        Network object used to store

    Notes
    -----
    The network is composed of a square grid of pyramidal cells, arranged in
    two layers (L5 and L2). The default in-plane separation of the grid points
    is 1.0 um, and the layer separation 1307.4 um. These can be adjusted after
    the net is created using the set_cell_positions-method. An all-to-all
    connectivity pattern is applied between cells. Inhibitory basket cells are
    present at a 1:3-ratio.

    This network was first described in Jones et al. 2009 [1]_ , and this code provides
    the implementation used in Neymotin et al. 2020 [2]_ .

    References
    ----------
    .. [1] Jones, Stephanie R., et al. "Quantitative Analysis and
           Biophysically Realistic Neural Modeling of the MEG Mu Rhythm:
           Rhythmogenesis and Modulation of Sensory-Evoked Responses."
           Journal of Neurophysiology 102, 3554–3572 (2009).
           https://doi.org/10.1152/jn.00535.2009

    .. [2] Neymotin, Samuel A, et al. 2020. "Human Neocortical Neurosolver (HNN), a New
           Software Tool for Interpreting the Cellular and Network Origin of Human
           MEG/EEG Data." eLife 9 (January):e51214. https://doi.org/10.7554/eLife.51214

    """
    hnn_core_root = op.dirname(hnn_core.__file__)
    if params is None:
        params = op.join(hnn_core_root, "param", "default.json")
    if isinstance(params, str):
        params = read_params(params)

    # Define cell types for Jones 2009 model
    # data is here in metaData format
    cell_types = {
        "L2_basket": {
            "cell_object": basket(cell_name="L2_basket"),
            "cell_metadata": {
                "morpho_type": "basket",
                "electro_type": "inhibitory",
                "layer": "2",
                "measure_dipole": False,
                "reference": "https://doi.org/10.7554/eLife.51214",
            },
        },
        "L2_pyramidal": {
            "cell_object": pyramidal(cell_name="L2_pyramidal"),
            "cell_metadata": {
                "morpho_type": "pyramidal",
                "electro_type": "excitatory",
                "layer": "2",
                "measure_dipole": True,
                "reference": "https://doi.org/10.7554/eLife.51214",
            },
        },
        "L5_basket": {
            "cell_object": basket(cell_name="L5_basket"),
            "cell_metadata": {
                "morpho_type": "basket",
                "electro_type": "inhibitory",
                "layer": "5",
                "measure_dipole": False,
                "reference": "https://doi.org/10.7554/eLife.51214",
            },
        },
        "L5_pyramidal": {
            "cell_object": pyramidal(cell_name="L5_pyramidal"),
            "cell_metadata": {
                "morpho_type": "pyramidal",
                "electro_type": "excitatory",
                "layer": "5",
                "measure_dipole": True,
                "reference": "https://doi.org/10.7554/eLife.51214",
            },
        },
    }

    # Create layer positions
    layer_dict = _create_cell_coords(
        n_pyr_x=mesh_shape[0],
        n_pyr_y=mesh_shape[1],
        z_coord=1307.4,  # Default layer separation
        inplane_distance=1.0,  # Default in-plane distance
    )

    # Map cell types to layer positions
    pos_dict = {
        "L5_pyramidal": layer_dict["L5_bottom"],
        "L2_pyramidal": layer_dict["L2_bottom"],
        "L5_basket": layer_dict["L5_mid"],
        "L2_basket": layer_dict["L2_mid"],
        "origin": layer_dict["origin"],
    }

    # Create network with cell types and positions
    net = Network(
        params,
        add_drives_from_params=add_drives_from_params,
        legacy_mode=legacy_mode,
        mesh_shape=mesh_shape,
        pos_dict=pos_dict,
        cell_types=cell_types,
    )

    delay = net.delay

    # source of synapse is always at soma

    # layer2 Pyr -> layer2 Pyr
    # layer5 Pyr -> layer5 Pyr
    lamtha = 3.0
    loc = "proximal"
    for target_cell in ["L2_pyramidal", "L5_pyramidal"]:
        for receptor in ["nmda", "ampa"]:
            key = (
                f"gbar_{_short_name(target_cell)}_{_short_name(target_cell)}_{receptor}"
            )
            weight = net._params[key]
            net.add_connection(
                target_cell,
                target_cell,
                loc,
                receptor,
                weight,
                delay,
                lamtha,
                allow_autapses=False,
            )

    # layer2 Basket -> layer2 Pyr
    src_cell = "L2_basket"
    target_cell = "L2_pyramidal"
    lamtha = 50.0
    loc = "soma"
    for receptor in ["gabaa", "gabab"]:
        key = f"gbar_L2Basket_L2Pyr_{receptor}"
        weight = net._params[key]
        net.add_connection(src_cell, target_cell, loc, receptor, weight, delay, lamtha)

    # layer5 Basket -> layer5 Pyr
    src_cell = "L5_basket"
    target_cell = "L5_pyramidal"
    lamtha = 70.0
    loc = "soma"
    for receptor in ["gabaa", "gabab"]:
        key = f"gbar_L5Basket_{_short_name(target_cell)}_{receptor}"
        weight = net._params[key]
        net.add_connection(src_cell, target_cell, loc, receptor, weight, delay, lamtha)

    # layer2 Pyr -> layer5 Pyr
    src_cell = "L2_pyramidal"
    lamtha = 3.0
    receptor = "ampa"
    for loc in ["proximal", "distal"]:
        key = f"gbar_L2Pyr_{_short_name(target_cell)}"
        weight = net._params[key]
        net.add_connection(src_cell, target_cell, loc, receptor, weight, delay, lamtha)

    # layer2 Basket -> layer5 Pyr
    src_cell = "L2_basket"
    lamtha = 50.0
    key = f"gbar_L2Basket_{_short_name(target_cell)}"
    weight = net._params[key]
    loc = "distal"
    receptor = "gabaa"
    net.add_connection(src_cell, target_cell, loc, receptor, weight, delay, lamtha)

    # xx -> layer2 Basket
    src_cell = "L2_pyramidal"
    target_cell = "L2_basket"
    lamtha = 3.0
    key = f"gbar_L2Pyr_{_short_name(target_cell)}"
    weight = net._params[key]
    loc = "soma"
    receptor = "ampa"
    net.add_connection(src_cell, target_cell, loc, receptor, weight, delay, lamtha)

    src_cell = "L2_basket"
    lamtha = 20.0
    key = f"gbar_L2Basket_{_short_name(target_cell)}"
    weight = net._params[key]
    loc = "soma"
    receptor = "gabaa"
    net.add_connection(src_cell, target_cell, loc, receptor, weight, delay, lamtha)

    # xx -> layer5 Basket
    src_cell = "L5_basket"
    target_cell = "L5_basket"
    lamtha = 20.0
    loc = "soma"
    receptor = "gabaa"
    key = f"gbar_L5Basket_{_short_name(target_cell)}"
    weight = net._params[key]
    net.add_connection(
        src_cell,
        target_cell,
        loc,
        receptor,
        weight,
        delay,
        lamtha,
        allow_autapses=False,
    )

    src_cell = "L5_pyramidal"
    lamtha = 3.0
    key = f"gbar_L5Pyr_{_short_name(target_cell)}"
    weight = net._params[key]
    loc = "soma"
    receptor = "ampa"
    net.add_connection(src_cell, target_cell, loc, receptor, weight, delay, lamtha)

    src_cell = "L2_pyramidal"
    lamtha = 3.0
    key = f"gbar_L2Pyr_{_short_name(target_cell)}"
    weight = net._params[key]
    loc = "soma"
    receptor = "ampa"
    net.add_connection(src_cell, target_cell, loc, receptor, weight, delay, lamtha)

    return net


def law_2021_model(
    params=None, add_drives_from_params=False, legacy_mode=False, mesh_shape=(10, 10)
):
    """Instantiate the expansion of Jones 2009 model to study beta
    modulated ERPs as described in
    Law et al. Cereb. Cortex 2021 [1]_

    Returns
    -------
    net : Instance of Network object
        Network object used to store the model used in
        Law et al. 2021.

    See Also
    --------
    jones_2009_model

    Notes
    -----
    Model reproduces results from Law et al. 2021
    This model differs from the default network model in several
    parameters including
    1) Increased GABAb time constants on L2/L5 pyramidal cells
    2) Decrease L5_pyramidal -> L5_pyramidal nmda weight
    3) Modified L5_basket -> L5_pyramidal inhibition weights
    4) Removal of L5 pyramidal somatic and basal dendrite calcium channels
    5) Replace L2_basket -> L5_pyramidal GABAa connection with GABAb
    6) Addition of L5_basket -> L5_pyramidal distal connection

    References
    ----------
    .. [1] Law, Robert G., et al. "Thalamocortical Mechanisms Regulating the
           Relationship between Transient Beta Events and Human Tactile
           Perception." Cerebral Cortex, 32, 668–688 (2022).
    """

    net = jones_2009_model(
        params, add_drives_from_params, legacy_mode, mesh_shape=mesh_shape
    )

    # Update biophysics (increase gabab duration of inhibition)
    net.cell_types["L2_pyramidal"]["cell_object"].synapses["gabab"]["tau1"] = 45.0
    net.cell_types["L2_pyramidal"]["cell_object"].synapses["gabab"]["tau2"] = 200.0
    net.cell_types["L5_pyramidal"]["cell_object"].synapses["gabab"]["tau1"] = 45.0
    net.cell_types["L5_pyramidal"]["cell_object"].synapses["gabab"]["tau2"] = 200.0

    # Decrease L5_pyramidal -> L5_pyramidal nmda weight
    net.connectivity[2]["nc_dict"]["A_weight"] = 0.0004

    # Modify L5_basket -> L5_pyramidal inhibition
    net.connectivity[6]["nc_dict"]["A_weight"] = 0.02  # gabaa
    net.connectivity[7]["nc_dict"]["A_weight"] = 0.005  # gabab

    # Remove L5 pyramidal somatic and basal dendrite calcium channels
    for sec in ["soma", "basal_1", "basal_2", "basal_3"]:
        del net.cell_types["L5_pyramidal"]["cell_object"].sections[sec].mechs["ca"]

    # Remove L2_basket -> L5_pyramidal gabaa connection
    del net.connectivity[10]  # Original paper simply sets gbar to 0.0

    # Add L2_basket -> L5_pyramidal gabab connection
    delay = net.delay
    src_cell = "L2_basket"
    target_cell = "L5_pyramidal"
    lamtha = 50.0
    weight = 0.0002
    loc = "distal"
    receptor = "gabab"
    net.add_connection(src_cell, target_cell, loc, receptor, weight, delay, lamtha)

    # Add L5_basket -> L5_pyramidal distal connection
    # ("Martinotti-like recurrent tuft connection")
    src_cell = "L5_basket"
    target_cell = "L5_pyramidal"
    lamtha = 70.0
    loc = "distal"
    receptor = "gabaa"
    key = f"gbar_L5Basket_L5Pyr_{receptor}"
    weight = net._params[key]
    net.add_connection(src_cell, target_cell, loc, receptor, weight, delay, lamtha)

    return net


# Remove params argument after updating examples
# (only relevant for Jones 2009 model)
def calcium_model(
    params=None, add_drives_from_params=False, legacy_mode=False, mesh_shape=(10, 10)
):
    """Instantiate the Jones 2009 model with improved calcium dynamics in
    L5 pyramidal neurons. For more details on changes to calcium dynamics
    see Kohl et al. Brain Topragr 2022 [1]_

    Returns
    -------
    net : Instance of Network object
        Network object used to store the Jones 2009 model with an improved
        calcium channel distribution.

    See Also
    --------
    jones_2009_model

    Notes
    -----
    This model builds on the Jones 2009 model by using a more biologically
    accurate distribution of calcium channels on L5 pyramidal cells.
    Specifically, this model introduces a distance dependent maximum
    conductance (gbar) on calcium channels such that the gbar linearly
    decreases along the dendrites in the direction of the soma.

    References
    ----------
    .. [1] Kohl, Carmen, et al. "Neural Mechanisms Underlying Human Auditory
           Evoked Responses Revealed By Human Neocortical Neurosolver."
           Brain Topography, 35, 19–35 (2022).
    """
    hnn_core_root = op.dirname(hnn_core.__file__)
    params_fname = op.join(hnn_core_root, "param", "default.json")
    if params is None:
        params = read_params(params_fname)

    net = jones_2009_model(
        params, add_drives_from_params, legacy_mode, mesh_shape=mesh_shape
    )

    # Replace L5 pyramidal cell template with updated calcium
    cell_name = "L5_pyramidal"
    pos = net.cell_types[cell_name]["cell_object"].pos
    net.cell_types[cell_name]["cell_object"] = pyramidal_ca(
        cell_name=cell_name, pos=pos
    )

    return net


def add_erp_drives_to_jones_model(net, tstart=0.0):
    """Add drives necessary for an event related potential (ERP)

    Parameters
    ----------
    net : Instance of Network object
        Network object that will be updated with ERP drives.
        Drives are updated in place.
    tstart : float | int
        Start time of sensory input in ms. (Default 0.0 ms)

    Notes
    -----
    The first proximal input arrives at cortex ~20 ms after sensory
    stimulus. The exact delay depends random number generator due to
    random sampling of times from a gaussian.
    """
    _validate_type(net, Network, "net", "Network")
    _validate_type(tstart, (float, int), "tstart", "float or int")

    # Add distal drive
    weights_ampa_d1 = {
        "L2_basket": 0.006562,
        "L2_pyramidal": 7e-6,
        "L5_pyramidal": 0.142300,
    }
    weights_nmda_d1 = {
        "L2_basket": 0.019482,
        "L2_pyramidal": 0.004317,
        "L5_pyramidal": 0.080074,
    }
    synaptic_delays_d1 = {"L2_basket": 0.1, "L2_pyramidal": 0.1, "L5_pyramidal": 0.1}
    net.add_evoked_drive(
        "evdist1",
        mu=63.53 + tstart,
        sigma=3.85,
        numspikes=1,
        weights_ampa=weights_ampa_d1,
        weights_nmda=weights_nmda_d1,
        location="distal",
        synaptic_delays=synaptic_delays_d1,
        event_seed=274,
    )

    # Add proximal drives
    weights_ampa_p1 = {
        "L2_basket": 0.08831,
        "L2_pyramidal": 0.01525,
        "L5_basket": 0.19934,
        "L5_pyramidal": 0.00865,
    }
    synaptic_delays_prox = {
        "L2_basket": 0.1,
        "L2_pyramidal": 0.1,
        "L5_basket": 1.0,
        "L5_pyramidal": 1.0,
    }
    net.add_evoked_drive(
        "evprox1",
        mu=26.61 + tstart,
        sigma=2.47,
        numspikes=1,
        weights_ampa=weights_ampa_p1,
        weights_nmda=None,
        location="proximal",
        synaptic_delays=synaptic_delays_prox,
        event_seed=544,
    )

    weights_ampa_p2 = {
        "L2_basket": 0.000003,
        "L2_pyramidal": 1.438840,
        "L5_basket": 0.008958,
        "L5_pyramidal": 0.684013,
    }
    net.add_evoked_drive(
        "evprox2",
        mu=137.12 + tstart,
        sigma=8.33,
        numspikes=1,
        weights_ampa=weights_ampa_p2,
        location="proximal",
        synaptic_delays=synaptic_delays_prox,
        event_seed=814,
    )



"""Batch Simulation."""

# Authors: Abdul Samad Siddiqui <abdulsamadsid1@gmail.com>
#          Nick Tolley <nicholas_tolley@brown.edu>
#          Ryan Thorpe <ryan_thorpe@brown.edu>
#          Mainak Jas <mjas@mgh.harvard.edu>

import numpy as np
import os
from joblib import Parallel, delayed, parallel_config

from .network import Network
from .externals.mne import _validate_type, _check_option
from .dipole import simulate_dipole
from .network_models import jones_2009_model


class BatchSimulate(object):
    """The BatchSimulate class.

    Parameters
    ----------
    set_params : func
        User-defined function that sets parameters in network drives.

            ``set_params(net, params) -> None``

        where ``net`` is a Network object and ``params`` is a dictionary
        of the parameters that will be set inside the function.
    net : Network object, optional
        The network model to use for simulations. Examples include the
        returned value of the following functions:

        - `jones_2009_model`: A network model based on Jones et al. (2009).
        - `law_2021_model`: A network model based on Law et al. (2021).
        - `calcium_model`: A network model incorporating calcium dynamics.

        Default is ``jones_2009_model()``.
    tstop : float, optional
        The stop time for the simulation. Default is 170 ms.
    dt : float, optional
        The time step for the simulation. Default is 0.025 ms.
    n_trials : int, optional
        The number of trials for the simulation. Default is 1.
    save_folder : str, optional
        The path to save the simulation outputs.
        Default is './sim_results'.
    batch_size : int, optional
        The maximum number of simulations saved in a single file.
        Default is 100.
    overwrite : bool, optional
        Whether to overwrite existing files and create file paths
        if they do not exist. Default is True.
    save_outputs : bool, optional
        Whether to save the simulation outputs to files. Default is False.
    save_dpl : bool, optional
        If True, save dipole results. Note, `save_outputs` must be True.
        Default: True.
    save_spiking : bool, optional
        If True, save spiking results. Note, `save_outputs` must be True.
        Default: False.
    save_lfp : bool, optional
        If True, save local field potential (lfp) results.
        Note, `save_outputs` must be True.
        Default: False.
    save_voltages : bool, optional
        If True, save voltages results. Note, `save_outputs` must be True.
        Default: False.
    save_currents : bool, optional
        If True, save currents results. Note, `save_outputs` must be True.
        Default: False.
    save_calcium : bool, optional
        If True, save calcium concentrations.
        Note, `save_outputs` must be True.
        Default: False.
    record_vsec : {False, 'all', 'soma'}
        Option to record voltages from all sections ('all'), or just
        the soma ('soma'). Default: False.
    record_isec : {False, 'all', 'soma'}
        Option to record voltages from all sections ('all'), or just
        the soma ('soma'). Default: False.
    postproc : bool, optional
        If True, smoothing (``dipole_smooth_win``) and scaling
        (``dipole_scalefctr``) values are read from the parameter file, and
        applied to the dipole objects before returning.
        Default: False.
    clear_cache : bool, optional
        Whether to clear the results cache after saving each batch.
        Default is False.
    summary_func : func, optional
        A function to calculate summary statistics from the simulation
        results. Default is None.

    Notes
    -----
    When ``save_output=True``, the saved files will appear as
    ``sim_run_{start_idx}-{end_idx}.npz`` in the specified `save_folder`
    directory. The `start_idx` and `end_idx` indicate the range of
    simulation indices contained in each file. Each file will contain
    a maximum of `batch_size` simulations, split evenly among the
    available files. If ``overwrite=True``, existing files with the same name
    will be overwritten.
    """

    def __init__(
        self,
        set_params,
        net=jones_2009_model(),
        tstop=170,
        dt=0.025,
        n_trials=1,
        save_folder="./sim_results",
        batch_size=100,
        overwrite=True,
        save_outputs=False,
        save_dpl=True,
        save_spiking=False,
        save_lfp=False,
        save_voltages=False,
        save_currents=False,
        save_calcium=False,
        record_vsec=False,
        record_isec=False,
        postproc=False,
        clear_cache=False,
        summary_func=None,
    ):
        _validate_type(net, Network, "net", "Network")
        _validate_type(tstop, types="numeric", item_name="tstop")
        _validate_type(dt, types="numeric", item_name="dt")
        _validate_type(n_trials, types="int", item_name="n_trials")
        _validate_type(save_folder, types="path-like", item_name="save_folder")
        _validate_type(batch_size, types="int", item_name="batch_size")
        _validate_type(save_outputs, types=(bool,), item_name="save_outputs")
        _validate_type(save_dpl, types=(bool,), item_name="save_dpl")
        _validate_type(save_spiking, types=(bool,), item_name="save_spiking")
        _validate_type(save_lfp, types=(bool,), item_name="save_lfp")
        _validate_type(save_voltages, types=(bool,), item_name="save_voltages")
        _validate_type(save_currents, types=(bool,), item_name="save_currents")
        _validate_type(save_calcium, types=(bool,), item_name="save_calcium")
        _check_option("record_vsec", record_vsec, ["all", "soma", False])
        _check_option("record_isec", record_isec, ["all", "soma", False])
        _validate_type(clear_cache, types=(bool,), item_name="clear_cache")

        if set_params is not None and not callable(set_params):
            raise TypeError("set_params must be a callable function")

        if summary_func is not None and not callable(summary_func):
            raise TypeError("summary_func must be a callable function")

        self.set_params = set_params
        self.net = net
        self.tstop = tstop
        self.dt = dt
        self.n_trials = n_trials
        self.save_folder = save_folder
        self.batch_size = batch_size
        self.overwrite = overwrite
        self.save_outputs = save_outputs
        self.save_dpl = save_dpl
        self.save_spiking = save_spiking
        self.save_lfp = save_lfp
        self.save_voltages = save_voltages
        self.save_currents = save_currents
        self.save_calcium = save_calcium
        self.record_vsec = record_vsec
        self.record_isec = record_isec
        self.postproc = postproc
        self.clear_cache = clear_cache
        self.summary_func = summary_func

    def run(
        self,
        param_grid,
        return_output=True,
        combinations=True,
        n_jobs=1,
        backend="loky",
        verbose=50,
    ):
        """Run batch simulations.

        Parameters
        ----------
        param_grid : dict
            Dictionary with parameter names and ranges.
        return_output : bool, optional
            Whether to return the simulation outputs. Default is True.
        combinations : bool, optional
            Whether to generate the Cartesian product of the parameter ranges.
            If False, generate combinations based on corresponding indices.
            Default is True.
        n_jobs : int, optional
            Number of parallel jobs. Default is 1.
        backend : str or joblib.parallel.ParallelBackendBase instance, optional
            The parallel backend to use. Can be one of `loky`, `threading`,
            `multiprocessing`, or `dask`. WARNING: currently only `loky` is
            completely operationable; all other backends are in
            development. Default is `loky`.
        verbose : int, optional
            The verbosity level for parallel execution. Default is 50.

        Returns
        -------
        results : dict
            Dictionary containing 'summary_statistics' and optionally
            'simulated_data'. 'simulated_data' may include keys: 'dpl', 'lfp',
            'spikes', 'voltages', 'param_values', 'net', 'times'.

        Notes
        -----
        Return content depends on `summary_func`, `return_output`, and
        `clear_cache` settings.
        """
        _validate_type(param_grid, types=(dict,), item_name="param_grid")
        _validate_type(n_jobs, types="int", item_name="n_jobs")
        _check_option(
            "backend", backend, ["loky", "threading", "multiprocessing", "dask"]
        )
        _validate_type(verbose, types="int", item_name="verbose")

        param_combinations = self._generate_param_combinations(param_grid, combinations)
        total_sims = len(param_combinations)
        batch_size = min(self.batch_size, total_sims)

        results = []
        simulated_data = []
        for i in range(0, total_sims, batch_size):
            start_idx = i
            end_idx = min(i + batch_size, total_sims)
            batch_results = self.simulate_batch(
                param_combinations[start_idx:end_idx],
                n_jobs=n_jobs,
                backend=backend,
                verbose=verbose,
            )

            if self.save_outputs:
                self._save(batch_results, start_idx, end_idx)

            if self.summary_func is not None:
                summary_statistics = self.summary_func(batch_results)
                results.append(summary_statistics)
                if not self.clear_cache:
                    simulated_data.append(batch_results)

            elif return_output and not self.clear_cache:
                simulated_data.append(batch_results)

            if self.clear_cache:
                del batch_results

        if return_output:
            if self.clear_cache:
                return {"summary_statistics": results}
            else:
                return {"summary_statistics": results, "simulated_data": simulated_data}

    def simulate_batch(
        self,
        param_combinations,
        n_jobs=1,
        backend="loky",
        verbose=50,
    ):
        """Simulate a batch of parameter sets in parallel.

        Parameters
        ----------
        param_combinations : list
            List of parameter combinations.
        n_jobs : int, optional
            Number of parallel jobs. Default is 1.
        backend : str or joblib.parallel.ParallelBackendBase instance, optional
            The parallel backend to use. Can be one of `loky`, `threading`,
            `multiprocessing`, or `dask`. WARNING: currently only `loky` is
            completely operationable; all other backends are in
            development. Default is `loky`.
        verbose : int, optional
            The verbosity level for parallel execution. Default is 50.

        Returns
        -------
        res: list
            List of dictionaries containing simulation results.
            Each dictionary contains the following keys along with their
            associated values:

            - `net`: The network model used for the simulation.
            - `dpl`: The simulated dipole.
            - `param_values`: The parameter values used for the simulation.
        """
        _validate_type(
            param_combinations, types=(list,), item_name="param_combinations"
        )
        _validate_type(n_jobs, types="int", item_name="n_jobs")
        _check_option(
            "backend", backend, ["loky", "threading", "multiprocessing", "dask"]
        )
        _validate_type(verbose, types="int", item_name="verbose")

        with parallel_config(backend=backend):
            res = Parallel(n_jobs=n_jobs, verbose=verbose)(
                delayed(self._run_single_sim)(params) for params in param_combinations
            )
        return res

    def _run_single_sim(self, param_values):
        """Run a single simulation.

        Parameters
        ----------
        param_values : dict
            Dictionary of parameter values.

        Returns
        -------
        dict
            Dictionary containing the simulation results.
            The dictionary contains the following keys:
            - `net`: The network model used for the simulation.
            - `dpl`: The simulated dipole.
            - `param_values`: The parameter values used for the simulation.
        """

        net = self.net.copy()
        self.set_params(param_values, net)

        results = {"net": net, "param_values": param_values}

        if self.save_dpl:
            dpl = simulate_dipole(
                net,
                tstop=self.tstop,
                dt=self.dt,
                n_trials=self.n_trials,
                record_vsec=self.record_vsec,
                record_isec=self.record_isec,
                postproc=self.postproc,
            )
            results["dpl"] = dpl

        if self.save_spiking:
            results["spiking"] = {
                "spike_times": net.cell_response.spike_times,
                "spike_types": net.cell_response.spike_types,
                "spike_gids": net.cell_response.spike_gids,
            }

        if self.save_lfp:
            results["lfp"] = net.rec_arrays

        if self.save_voltages:
            results["voltages"] = net.cell_response.vsec

        if self.save_currents:
            results["currents"] = net.cell_response.isec

        if self.save_calcium:
            results["calcium"] = net.cell_response.ca

        return results

    def _generate_param_combinations(self, param_grid, combinations=True):
        """Generate combinations of parameters from the grid.

        Parameters
        ----------
        param_grid : dict
            Dictionary with parameter names and ranges.
        combinations : bool, optional
            Whether to generate the Cartesian product of the parameter ranges.
            If False, generate combinations based on corresponding indices.
            Default is True.

        Returns
        -------
        param_combinations: list
            List of parameter combinations.
        """
        from itertools import product

        keys, values = zip(*param_grid.items())
        if combinations:
            param_combinations = [
                dict(zip(keys, combination)) for combination in product(*values)
            ]
        else:
            param_combinations = [
                dict(zip(keys, combination)) for combination in zip(*values)
            ]
        return param_combinations

    def _save(self, results, start_idx, end_idx):
        """Save simulation results to files.

        Parameters
        ----------
        results : list
            The results to save.
        start_idx : int
            The index of the first simulation in this batch.
        end_idx : int
            The index of the last simulation in this batch.
        """
        _validate_type(results, types=(list,), item_name="results")
        _validate_type(start_idx, types="int", item_name="start_idx")
        _validate_type(end_idx, types="int", item_name="end_idx")

        if not os.path.exists(self.save_folder):
            os.makedirs(self.save_folder)

        save_data = {"param_values": [result["param_values"] for result in results]}

        attributes_to_save = [
            "dpl",
            "spiking",
            "lfp",
            "voltages",
            "currents",
            "calcium",
        ]
        for attr in attributes_to_save:
            if getattr(self, f"save_{attr}") and attr in results[0]:
                save_data[attr] = [result[attr] for result in results]

        metadata = {
            "batch_size": self.batch_size,
            "n_trials": self.n_trials,
            "tstop": self.tstop,
            "dt": self.dt,
        }
        save_data["metadata"] = metadata

        file_name = os.path.join(self.save_folder, f"sim_run_{start_idx}-{end_idx}.npz")
        if os.path.exists(file_name) and not self.overwrite:
            raise FileExistsError(
                f"File {file_name} already exists and overwrite is set to False."
            )

        np.savez(file_name, **save_data)

    def load_results(self, file_path, return_data=None):
        """Load simulation results from a file.

        Parameters
        ----------
        file_path : str
            The path to the file containing the simulation results.
        return_data : list of str, optional
            List of data types to return. If None, returns the types specified
            during initialization. Defaults to None.

        Returns
        -------
        results : dict
            Dictionary containing the loaded results and parameter values.
        """
        if return_data is None:
            return_data = []
            if self.save_dpl:
                return_data.append("dpl")
            if self.save_spiking:
                return_data.append("spiking")
            if self.save_lfp:
                return_data.append("lfp")
            if self.save_voltages:
                return_data.append("voltages")
            if self.save_currents:
                return_data.append("currents")
            if self.save_calcium:
                return_data.append("calcium")

        data = np.load(file_path, allow_pickle=True)
        results = {
            key: data[key].tolist()
            for key in data.files
            if key in return_data or key == "param_values"
        }
        return results

    def load_all_results(self):
        """Load all simulation results from the files in `self.save_folder`.

        Returns
        -------
        all_results : list
            List of dictionaries containing all loaded simulation results.
        """
        all_results = []
        for file_name in os.listdir(self.save_folder):
            if file_name.startswith("sim_run_") and file_name.endswith(".npz"):
                file_path = os.path.join(self.save_folder, file_name)
                results = self.load_results(file_path)
                all_results.append(results)
        return all_results



# Authors: Mainak Jas <mainakjas@gmail.com>

from glob import glob

import matplotlib.pyplot as plt
import pytest
import numpy as np

from hnn_core import CellResponse, read_spikes


def test_cell_response(tmp_path):
    """Test CellResponse object."""
    # Round-trip test
    spike_times = [[2.3456, 7.89], [4.2812, 93.2]]
    spike_gids = [[1, 3], [5, 7]]
    spike_types = [["L2_pyramidal", "L2_basket"], ["L5_pyramidal", "L5_basket"]]
    tstart, tstop, fs = 0.1, 98.4, 1000.0
    sim_times = np.arange(tstart, tstop, 1 / fs)
    gid_ranges = {
        "L2_pyramidal": range(1, 2),
        "L2_basket": range(3, 4),
        "L5_pyramidal": range(5, 6),
        "L5_basket": range(7, 8),
    }
    default_cell_type_names = ["L2_basket", "L2_pyramidal", "L5_basket", "L5_pyramidal"]
    cell_response = CellResponse(
        cell_type_names=default_cell_type_names,
        spike_times=spike_times,
        spike_gids=spike_gids,
        spike_types=spike_types,
        times=sim_times,
    )

    assert set(cell_response.cell_types) == set(gid_ranges.keys())
    assert cell_response.spike_times_by_type["L2_basket"] == [[7.89], []]
    assert cell_response.spike_times_by_type["L5_pyramidal"] == [[], [4.2812]]

    kwargs_hist = dict(alpha=0.25)
    fig = cell_response.plot_spikes_hist(show=False, **kwargs_hist)
    assert all(
        patch.get_alpha() == kwargs_hist["alpha"] for patch in fig.axes[0].patches
    ), "Alpha value not applied to all patches"

    # Testing writing using txt files
    cell_response.write(tmp_path / "spk_%d.txt")

    # Testing reading from txt files
    assert cell_response == read_spikes(tmp_path / "spk_*.txt")

    assert "CellResponse | 2 simulation trials" in repr(cell_response)

    # reset clears all recorded variables, but leaves simulation time intact
    assert len(cell_response.times) == len(sim_times)
    sim_attributes = [
        "_spike_times",
        "_spike_gids",
        "_spike_types",
        "_vsec",
        "_isec",
        "_ca",
    ]
    net_attributes = ["_times", "_cell_type_names"]  # `Network.__init__`
    # creates these check that we always know which response attributes are
    # simulated see #291 for discussion; objective is to keep cell_response
    # size small
    assert sorted(list(cell_response.__dict__.keys())) == sorted(
        sim_attributes + net_attributes
    )

    # Test recovery of empty spike files
    empty_spike = CellResponse(
        cell_type_names=default_cell_type_names,
        spike_times=[[], []],
        spike_gids=[[], []],
        spike_types=[[], []],
    )
    empty_spike.write(tmp_path / "empty_spk_%d.txt")
    empty_spike.write(tmp_path / "empty_spk.txt")
    empty_spike.write(tmp_path / "empty_spk_{0}.txt")
    assert empty_spike == read_spikes(tmp_path / "empty_spk_*.txt")

    assert "CellResponse | 2 simulation trials" in repr(empty_spike)

    with pytest.raises(TypeError, match="spike_times should be a list of lists"):
        cell_response = CellResponse(
            cell_type_names=default_cell_type_names,
            spike_times=([2.3456, 7.89], [4.2812, 93.2]),
            spike_gids=spike_gids,
            spike_types=spike_types,
        )

    with pytest.raises(TypeError, match="spike_times should be a list of lists"):
        cell_response = CellResponse(
            cell_type_names=default_cell_type_names,
            spike_times=[1, 2],
            spike_gids=spike_gids,
            spike_types=spike_types,
        )

    with pytest.raises(
        ValueError,
        match="spike times, gids, and types should be lists of the same length",
    ):
        cell_response = CellResponse(
            cell_type_names=default_cell_type_names,
            spike_times=[[2.3456, 7.89]],
            spike_gids=spike_gids,
            spike_types=spike_types,
        )

    cell_response = CellResponse(
        cell_type_names=default_cell_type_names,
        spike_times=spike_times,
        spike_gids=spike_gids,
        spike_types=spike_types,
    )

    with pytest.raises(
        TypeError, match="spike_types should be str, list, dict, or None"
    ):
        cell_response.plot_spikes_hist(spike_types=1, show=False)

    with pytest.raises(
        TypeError,
        match=r"spike_types\[ev\] must be a list\. "
        r"Got int\.",
    ):
        cell_response.plot_spikes_hist(spike_types={"ev": 1}, show=False)

    with pytest.raises(
        ValueError,
        match=r"Elements of spike_types must map to"
        r" mutually exclusive input types\. L2_basket is found"
        r" more than once\.",
    ):
        cell_response.plot_spikes_hist(
            spike_types={"ev": ["L2_basket", "L2_b"]}, show=False
        )

    with pytest.raises(ValueError, match="No input types found for ABC"):
        cell_response.plot_spikes_hist(spike_types="ABC", show=False)

    with pytest.raises(
        ValueError, match="tstart and tstop must be of type int or float"
    ):
        cell_response.mean_rates(tstart=0.1, tstop="ABC", gid_ranges=gid_ranges)

    with pytest.raises(ValueError, match="tstop must be greater than tstart"):
        cell_response.mean_rates(tstart=0.1, tstop=-1.0, gid_ranges=gid_ranges)

    with pytest.raises(
        ValueError,
        match="Invalid mean_type. Valid arguments include 'all', 'trial', or 'cell'.",
    ):
        cell_response.mean_rates(
            tstart=tstart, tstop=tstop, gid_ranges=gid_ranges, mean_type="ABC"
        )

    test_rate = (1 / (tstop - tstart)) * 1000

    assert cell_response.mean_rates(tstart, tstop, gid_ranges) == {
        "L5_pyramidal": test_rate / 2,
        "L5_basket": test_rate / 2,
        "L2_pyramidal": test_rate / 2,
        "L2_basket": test_rate / 2,
    }
    assert cell_response.mean_rates(tstart, tstop, gid_ranges, mean_type="trial") == {
        "L5_pyramidal": [0.0, test_rate],
        "L5_basket": [0.0, test_rate],
        "L2_pyramidal": [test_rate, 0.0],
        "L2_basket": [test_rate, 0.0],
    }
    assert cell_response.mean_rates(tstart, tstop, gid_ranges, mean_type="cell") == {
        "L5_pyramidal": [[0.0], [test_rate]],
        "L5_basket": [[0.0], [test_rate]],
        "L2_pyramidal": [[test_rate], [0.0]],
        "L2_basket": [[test_rate], [0.0]],
    }

    # Write spike file with no 'types' column
    for fname in sorted(glob(str(tmp_path / "spk_*.txt"))):
        times_gids_only = np.loadtxt(fname, dtype=str)[:, (0, 1)]
        np.savetxt(fname, times_gids_only, delimiter="\t", fmt="%s")

    # Check that spike_types are updated according to gid_ranges
    cell_response = read_spikes(tmp_path / "spk_*.txt", gid_ranges=gid_ranges)
    assert cell_response.spike_types == spike_types

    # Check for gid_ranges errors
    with pytest.raises(
        ValueError,
        match="gid_ranges must be provided if spike types are unspecified in the file ",
    ):
        cell_response = read_spikes(tmp_path / "spk_*.txt")
    with pytest.raises(
        ValueError, match="gid_ranges should contain only disjoint sets of gid values"
    ):
        gid_ranges = {
            "L2_pyramidal": range(3),
            "L2_basket": range(2, 4),
            "L5_pyramidal": range(4, 6),
            "L5_basket": range(6, 8),
        }
        cell_response = read_spikes(tmp_path / "spk_*.txt", gid_ranges=gid_ranges)
    plt.close("all")



"""
=======================================
04. Use MPI backend for parallelization
=======================================

This example demonstrates how to use the MPI backend for
simulating dipoles using HNN-core.

The MPI backend allows running the simulation in parallel across neurons in the
network even with a single trial. For this, you will
need the :ref:`MPI related software <install>` installed. Note that if you
want to simulate in parallel across trials, the Joblib backend allows this
without the need to install and configure MPI.
"""

# Authors: Mainak Jas <mjas@mgh.harvard.edu>
#          Blake Caldwell <blake_caldwell@brown.edu>

###############################################################################
# Let us import hnn_core
import os.path as op

import hnn_core
from hnn_core import simulate_dipole, jones_2009_model

###############################################################################
# Following :ref:`the alpha example
# <sphx_glr_auto_examples_plot_simulate_alpha.py>`, we add a
# ~10 Hz "bursty" drive starting at 50 ms and continuing to the end of the
# simulation. Each burst consists of a pair (2) of spikes, spaced 10 ms apart.
# The occurrence of each burst is jittered by a random, normally distributed
# amount (20 ms standard deviation). We repeat the burst train 10 times, each
# time with unique randomization.
net = jones_2009_model()

weights_ampa = {'L2_pyramidal': 5.4e-5, 'L5_pyramidal': 5.4e-5}
net.add_bursty_drive(
    'bursty', tstart=50., burst_rate=10, burst_std=20., numspikes=2,
    spike_isi=10, n_drive_cells=10, location='distal',
    weights_ampa=weights_ampa, event_seed=278)

###############################################################################
# Finally, to simulate we use the
# :class:`~hnn_core.parallel_backends.MPIBackend` class. This will
# start the simulation across the number of processors (cores) specified by
# ``n_procs`` using MPI. The ``'mpiexec'`` launcher is used from
# ``openmpi``, which must be installed on the system
from hnn_core import MPIBackend

with MPIBackend(n_procs=2, mpi_cmd='mpiexec'):
    dpls = simulate_dipole(net, tstop=310., n_trials=1)

trial_idx = 0
dpls[trial_idx].plot()
